# General utils

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pingouin as pg

from matplotlib.ticker import MaxNLocator
from sklearn.metrics import mutual_info_score


def compute_high_corr_pairs(df, corr_thres=0.7, method='pearson'):
    """
    Computes the correlation coefficient between every pair of variables and returns those pairs with an absolute value
    above the given threshold.
    *Note* all variables must be numerical.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    corr_thres : float, default=0.7
        Correlation theshold to consider two variables as strongly correlated.
    method : str, default='pearson'
        Method of correlation (pearson, kendall, spearman, or callable function -
        see https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.corr.html).

    Returns
    ----------
    num_pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables together with the correlation coefficient value.
    """
    corr_df = df.corr(method=method)
    # We only keep the pairs with a corr coefficient above the threshold and melt the DataFrame to have a row per pair
    num_pairs = corr_df.replace({1: 0})[np.abs(corr_df.replace({1: 0})) > corr_thres].reset_index()\
        .melt(id_vars='index', value_vars=corr_df.columns[1:]).dropna().sort_values('value', ascending=False) \
        .rename(columns={'index': 'var1', 'variable': 'var2'}).reset_index(drop=True)
    return num_pairs


def compute_highly_related_mixed_vars(df, num_vars, cat_vars, np2_thres=0.14):
    """
    Computes the dependency between pairs of numerical and categorical variables through partial eta squared, and
    returns those pairs with a value above the given threshold.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    num_vars : string, list, series, or vector array
        Numerical variable name(s).
    cat_vars : string, list, series, or vector array
        Categorical variable name(s).
    np2_thres : float, default=0.14
        Threshold to consider two variables as strongly related (see
        https://www.spss-tutorials.com/effect-size/#anova-partial-eta-squared).

    Returns
    ----------
    pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables together with the partial eta squared value.
    """
    cross_corr_df = cross_corr_ratio(df[cat_vars], df[num_vars])
    # We keep only the pairs with a np2 above the threshold and melt the notebook to have a row per pair
    pairs = cross_corr_df[cross_corr_df > np2_thres].reset_index()\
        .melt(id_vars='index', value_vars=cross_corr_df.columns[1:]) .dropna()\
        .sort_values('value', ascending=False).rename(columns={'index': 'var1', 'variable': 'var2'})\
        .reset_index(drop=True)

    # For every pair <cat_var, num_var>, we're interested in the relationship in both ways
    pairs = pd.concat([pairs, pairs.rename(columns={'var1': 'var2', 'var2': 'var1'})], ignore_index=True)
    pairs = pairs.sort_values(['value', 'var1', 'var2'], ascending=[False, True, True]).reset_index(drop=True)
    return pairs


def compute_highly_related_categorical_vars(df, mi_thres=0.6):
    """
    Computes the dependency between paris of categorical variables through mutual information
    (https://en.wikipedia.org/wiki/Mutual_information), and returns those pairs with a value above the given threshold.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    mi_thres : float, default=0.6
        Threshold to consider two variables as strongly related.

    Returns
    ----------
    cat_pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables together with the mutual information score.
    """
    data = []
    cat_vars = list(df.columns)
    for i in range(len(cat_vars) - 1):
        for j in range(i + 1, len(cat_vars)):
            df_f = df[(~df[cat_vars[i]].isnull()) & (~df[cat_vars[j]].isnull())]
            mis = mutual_info_score(df_f[cat_vars[i]], df_f[cat_vars[j]])
            if mis > mi_thres:
                data.append((cat_vars[i], cat_vars[j], mis))
                data.append((cat_vars[j], cat_vars[i], mis))

    cat_pairs = pd.DataFrame(data=data, columns=['var1', 'var2', 'value']).sort_values(
        ['value', 'var1', 'var2'], ascending=[False, True, True]).reset_index(drop=True)
    return cat_pairs


def cross_corr_ratio(df1, df2):
    """
    Calculates the correlation ratio of every column in df1 with every column in df2
    https://en.wikipedia.org/wiki/Correlation_ratio

    Parameters
    ----------
    df1 : `pandas.DataFrame`
    df2 : `pandas.DataFrame`

    Returns
    ----------
    corr_df: `pandas.DataFrame`
        DataFrame with the correlation ratio of every pair of columns from df1 and df2.
    """
    corr_coeffs = []
    df_aux = pd.concat([df1, df2], axis=1)
    for col in df1.columns:
        col_corr_coeffs = []
        for dv in df2.columns:
            col_corr_coeffs.append(pg.anova(data=df_aux, dv=dv, between=col)['np2'].iloc[0])
        corr_coeffs.append(col_corr_coeffs)

    corr_df = pd.DataFrame(corr_coeffs, index=df1.columns, columns=df2.columns).transpose()
    return corr_df


def get_axis(i, axs, ncols, nrows):
    ax = None
    if ncols > 1 and nrows > 1:
        ax = axs[i // ncols, i % ncols]
    elif ncols == 1 and nrows == 1:
        ax = axs
    else:
        ax = axs[i]
    return ax


def savefig(output_path=None, savefig_kws=None):
    if output_path is not None:
        if savefig_kws is not None:
            plt.savefig(output_path, **savefig_kws)
        else:
            plt.savefig(output_path, format='jpg', bbox_inches='tight', dpi=300)


def plot_optimal_normalized_elbow(values, kl, ax, optimal_label='', xlabel='', ylabel=''):
    n_components = len(values)

    ax.plot([np.nan] + list(kl.y_normalized), color='#332288', label='')
    ax.plot([np.nan] + list(kl.y_difference), color='#008695', label='Difference curve')

    ax.axvline(kl.knee, linestyle='--', linewidth=1, color='#E73F74', label=optimal_label)
    ax.set_xticks(np.append(ax.get_xticks()[1:-1], [kl.knee]))
    ax.set_xlim(-n_components * 0.02, n_components * 1.02)
    ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    ax.set_xlabel(xlabel, fontsize=13, labelpad=15)
    ax.set_ylabel(ylabel, fontsize=13, labelpad=8)
    ax.legend(fontsize=12, labelspacing=0.5)



"""Utils for clustering"""
# Author: Miguel Alvarez-Garcia

import numpy as np
import pandas as pd

from statsmodels.stats.weightstats import DescrStatsW


def weighted_sum_of_squared_distances(df, cluster_arr, weights=None):
    """
    Calculates the weighted sum of squared distances for given clusters

    Parameters
    ----------
    df : `pandas.DataFrame`
    cluster_arr : `numpy.array`
    weights : `numpy.array`, default=None

    Returns
    ----------
    wssd : float
        Weighted sum of squared distances
    """

    if weights is None:
        weights = np.ones(len(cluster_arr))

    cl_df = pd.DataFrame(data={'cluster': cluster_arr, 'weights': weights})
    cl_df.index = df.index

    wssd = 0
    for cl in cl_df['cluster'].unique():
        if cl_df[cl_df['cluster'] == cl].shape[0] > 1:
            local_weights = cl_df.loc[cl_df['cluster'] == cl, 'weights']
            descr = DescrStatsW(df.loc[cl_df[cl_df['cluster'] == cl].index],
                                local_weights)
            wssd += np.sum(np.array(list(map(np.sum, descr.demeaned ** 2))) * local_weights)

    return wssd



from .dimensionality_reduction import (
    DimensionalityReduction
)

from .viz_utils import (
    plot_compare_pca_based_components
)

__all__ = [
    # Dimensionality reduction main class
    "DimensionalityReduction",
    # visualization utils
    "plot_compare_pca_based_components"
]



"""Visualization utils for the classifier module"""
# Author: Miguel Alvarez-Garcia

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import shap
import utils

from matplotlib.collections import QuadMesh
from sklearn.preprocessing import label_binarize
from sklearn.metrics import roc_curve, auc
from ..utils import savefig

sns.set_style('whitegrid')


def plot_shap_importances(model, X, n_top=7, output_path=None, savefig_kws=None):
    """
    Plots shap importance values, calculated as the combined average of the absolute values of the shap values
    for all classes.

    Parameters
    ----------
    model : `scikit-learn.Estimator`
        Classification model (already trained).
    X : `pandas.DataFrame` or `numpy.ndarray`
        Observations (predictors).
    n_top : int, default=7
        Top n features to be displayed. The importances of the rest are aggregated and displayed under the tag "Rest".
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    si = utils.shap_importances(model, X)
    low_imp = si.loc[n_top:]
    si = si.loc[:n_top - 1]
    si = si.append(pd.DataFrame({'variable_name': ['Rest'], 'shap_importance': [low_imp['shap_importance'].sum()]}),
                   ignore_index=True)
    fig, ax = plt.subplots(figsize=(10, 0.675 * si.shape[0]))
    ax = sns.barplot(x='shap_importance', y='variable_name', data=si, color='#ff0051')
    ax.bar_label(ax.containers[0], padding=5, fmt='%.4f', fontsize=10)
    ax.set_xlabel('mean(|SHAP values|)', fontsize=12)
    ax.set_ylabel('')
    fig.tight_layout()
    savefig(output_path, savefig_kws)


def plot_shap_importances_beeswarm(model, X, class_id, n_top=10, output_path=None, savefig_kws=None):
    """
    Plots a summary of shap values for a specific class of the target variable. This uses shap beeswarm plot
    (https://shap.readthedocs.io/en/latest/example_notebooks/api_examples/plots/beeswarm.html).

    Parameters
    ----------
    model : `scikit-learn.Estimator`
        Classification model (already trained).
    X : `pandas.DataFrame` or `numpy.ndarray`
        Observations (predictors).
    class_id : int
        The class for which to show the SHAP values.
    n_top : int, default=7
        Top n features to be displayed. The importances of the rest are aggregated and displayed under the tag "Rest".
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    explainer = shap.Explainer(model)
    shap_values = explainer(X)
    shap.plots.beeswarm(shap_values[:, :, class_id], show=False, max_display=n_top+1)
    plt.title(f'SHAP values summary for class {class_id}', fontsize=13)
    savefig(output_path, savefig_kws)


def plot_confusion_matrix(cf, group_names=None, count=True, percent=True, sum_stats=True, xyticks=True,
                          xyplotlabels=True, figsize=None, cmap='Blues', title=None, output_path=None,
                          savefig_kws=None):
    """
    This function makes a pretty plot of an sklearn Confusion Matrix cf using a Seaborn heatmap visualization.

    Inspired by : https://github.com/DTrimarchi10/confusion_matrix/blob/master/cf_matrix.py

    Parameters
    ---------
    cf : `pandas.DataFrame`
        Confusion matrix to be passed in.
    group_names : list, default=None
        List of strings that represent the labels row by row to be shown in each square.
    count : boolean, default=True
        If True, show the raw number in the confusion matrix.
    percent : boolean, default=True
        If True, show the percentages in the confusion matrix.
    sum_stats : boolean, default=True
        If True, show precision and recall per class, and global accuracy, appended to the matrix.
    xyticks : boolean, default=True
        If True, show x and y ticks.
    xyplotlabels : boolean, default=True
        f True, show 'Observed values' and 'Predicted values' on the figure.
    figsize : tuple, default=None
        Tuple representing the figure size. Default will be the matplotlib default value.
    cmap : `matplotlib.pyplot.cm`, default='Blues'
        Colormap of the values displayed from matplotlib.pyplot.cm. Default is 'Blues'
        See http://matplotlib.org/examples/color/colormaps_reference.html
    title : str, feault=None
        Title for the heatmap.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """

    # CODE TO GENERATE TEXT INSIDE EACH SQUARE
    blanks = ['' for i in range(cf.size)]

    if group_names and len(group_names) == cf.size:
        group_labels = ["{}\n".format(value) for value in group_names]
    else:
        group_labels = blanks

    flatten = cf.values.flatten()
    if count:
        # Format of totals is different
        if sum_stats:
            group_counts = ["{0:0.0f}\n".format(flatten[i]) if (
                        i % cf.shape[0] < cf.shape[0] - 1 and i // cf.shape[0] < cf.shape[0] - 1)
                            else "{0:.2%}".format(flatten[i]) for i in range(len(flatten))]
        else:
            group_counts = ["{0:0.0f}\n".format(value) for value in flatten]
    else:
        group_counts = blanks

    if percent:
        # Format of totals is different
        if sum_stats:
            group_percentages = ["{0:.2%}".format(flatten[i] / np.sum(cf.values)) if (
                        i % cf.shape[0] < cf.shape[0] - 1 and i // cf.shape[0] < cf.shape[0] - 1)
                                 else "" for i in range(len(flatten))]
        else:
            group_percentages = ["{0:.2%}".format(value) for value in cf.values.flatten() / np.sum(cf.values)]
    else:
        group_percentages = blanks

    box_labels = [f"{v1}{v2}{v3}".strip() for v1, v2, v3 in zip(group_labels, group_counts, group_percentages)]
    box_labels = np.asarray(box_labels).reshape(cf.shape[0], cf.shape[1])

    # Cluster names as tick labels. If precision, recall, and global accuracy are included, no tick label is displayed
    # for them
    xyticklabels = False
    if xyticks:
        if sum_stats:
            xyticklabels = list(map(lambda x: x[1], cf.columns[:-1])) + ['']
        else:
            xyticklabels = list(map(lambda x: x[1], cf.columns))

    # MAKE THE HEATMAP VISUALIZATION
    ax, fig = plt.subplots(figsize=figsize)
    sns.heatmap(cf.values, annot=box_labels, fmt="", cmap=cmap, cbar=False, xticklabels=xyticklabels,
                yticklabels=xyticklabels)

    # Remove colors from totals
    quadmesh = ax.findobj(QuadMesh)[0]
    facecolors = quadmesh.get_facecolors()
    # make colors of the last column white
    facecolors[np.arange(cf.shape[0] - 1, cf.size, cf.shape[0])] = np.array([0, 0, 0, 0.05])
    facecolors[np.arange(cf.size - cf.shape[0], cf.size)] = np.array([0, 0, 0, 0.05])
    quadmesh.set_facecolors = facecolors

    if xyplotlabels:
        plt.ylabel('Observed values', fontsize=11)
        plt.xlabel('Predicted values', fontsize=11)

    if title:
        plt.title(title)

    plt.yticks(rotation=0)
    plt.tight_layout(pad=2)
    savefig(output_path, savefig_kws)


def plot_roc_curves(X, y, model, output_path=None, savefig_kws=None):
    """
    Plots ROC curve for every class.

    Parameters
    ---------
    X : `pandas.DataFrame` or `numpy.ndarray`
        Predictor values.
    y : `pandas.Series` or `numpy.array`
        Target values.
    model : `scikit-learn.Estimator`
        Classification model (already trained).
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    classes = np.sort(np.unique(y))
    y_score = model.predict_proba(X)
    y_test_b = label_binarize(y, classes=classes)

    # Compute ROC curve and AUC for every class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for i in range(len(classes)):
        fpr[i], tpr[i], _ = roc_curve(y_test_b[:, i], y_score[:, i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    # Compute micro-average ROC curve and ROC area
    fpr["micro"], tpr["micro"], _ = roc_curve(y_test_b.ravel(), y_score.ravel())
    roc_auc["micro"] = auc(fpr["micro"], tpr["micro"])

    # Plot curves
    ncols = 3
    if len(classes) % 3 > 0 and len(classes) % 2 == 0:
        ncols = 2
    nrows = int(np.ceil(len(classes) / ncols))

    fig, axs = plt.subplots(nrows, ncols, figsize=(4 * ncols, 3 * nrows))
    for i in range(len(classes)):
        axs[i // ncols, i % ncols].plot(fpr[i], tpr[i], color='#11A579',
                                        label='ROC curve (area = %0.4f)' % roc_auc[i])
        axs[i // ncols, i % ncols].plot([0, 1], [0, 1], color='#7F3C8D', linestyle='--')
        axs[i // ncols, i % ncols].set_xlim([-0.025, 1.025])
        axs[i // ncols, i % ncols].set_ylim([0.0, 1.05])
        axs[i // ncols, i % ncols].set_title(f'Cluster {classes[i]}', fontsize=12, pad=10)
        axs[i // ncols, i % ncols].legend(loc="lower right")

        if i // ncols == nrows - 1:
            axs[i // ncols, i % ncols].set_xlabel('False Positive Rate', fontsize=11, labelpad=10)
        if i % ncols == 0:
            axs[i // ncols, i % ncols].set_ylabel('True Positive Rate', fontsize=11, labelpad=10)

    fig.tight_layout()
    savefig(output_path, savefig_kws)



from unittest import TestCase


class TestPreprocessing(TestCase):
    def test_imputation(self):
        self.assertEqual(0, 0)

    def test_outliers(self):
        self.assertEqual(0, 0)


class TestDimReduction(TestCase):
    def test_dim_reduction(self):
        self.assertEqual(0, 0)


class TestClustering(TestCase):
    def test_clustering(self):
        self.assertEqual(0, 0)


class TestClassifier(TestCase):
    def test_classifier(self):
        self.assertEqual(0, 0)



"""Utils for visualization"""
# Author: Miguel Alvarez-Garcia

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from kneed import KneeLocator
from table_utils import compare_cluster_means_to_global_means
from ..utils import get_axis, plot_optimal_normalized_elbow, savefig


sns.set_style('whitegrid')

CARTO_COLORS = ['#7F3C8D', '#11A579', '#3969AC', '#F2B701', '#E73F74', '#80BA5A', '#E68310',  '#008695', '#CF1C90',
                '#f97b72', '#4b4b8f', '#A5AA99']


def plot_score_comparison(scores, cluster_range, metric_name='Weighted sum of squared distances', output_path=None,
                          savefig_kws=None):
    """
    Plots the comparison in performance between the different clustering algorithms.

    Parameters
    ----------
    scores : dict
        Dictionary <algorithm, list of scores>
    cluster_range : [min (int), max (int))
        Range of number of clusters computed. This will be displayed on the x-axis.
    metric_name : str, default='Weighted sum of squared distances'
        Name of the metric used for comparison. This will be displayed on the y-label.
        Default is 'Weighted sum of squared distances', which corresponds to inertia.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    plt.figure(figsize=(10, 5))
    i = 0
    for algorithm in scores:
        plt.plot(scores[algorithm], label=algorithm, color=CARTO_COLORS[i])
        i += 1

    plt.xlabel('Number of clusters', fontsize=12, labelpad=15)
    plt.ylabel(metric_name, fontsize=12, labelpad=15)
    plt.xticks(ticks=range(len(list(scores.values())[0])), labels=range(*cluster_range))
    plt.tight_layout()
    plt.legend(fontsize=12, title='Algorithm', title_fontsize=13, labelspacing=0.5)

    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_optimal_components_normalized(scores, max_clusters, metric_name, output_path=None, savefig_kws=None):
    """
    Plots the normalized curve used for computing the optimal number of clusters.

    Parameters
    ----------
    scores : dict
        Dictionary <algorithm, list of scores>
    max_clusters : int
        Maximum number of clusters allowed.
    metric_name : str, default='Weighted sum of squared distances'
        Name of the metric used for comparison. This will be displayed on the y-label.
        Default is 'Weighted sum of squared distances', which corresponds to inertia.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    fig, ax = plt.subplots(figsize=(8, 5))
    kl = KneeLocator(x=range(1, max_clusters + 1), y=scores, curve='convex', direction='decreasing')
    plot_optimal_normalized_elbow(scores, kl, ax, optimal_label='Optimal number of clusters',
                                  xlabel='Number of clusters', ylabel=f'Normalized {metric_name}')
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_clustercount(df, output_path=None, savefig_kws=None):
    """
    Plots a bar plot with cluster counts.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing at least a column named 'cluster_cat' with the cluster labels.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    plt.figure(figsize=(df['cluster_cat'].nunique(), 5))
    sns.countplot(x='cluster_cat', data=df, color='#332288', alpha=0.9, order=np.sort(df['cluster_cat'].unique()))
    # plt.xticks(rotation=30)
    plt.ylabel('count', fontsize=12, labelpad=10)
    plt.xlabel('clusters', fontsize=12, labelpad=10)
    plt.tight_layout(pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_cluster_means_to_global_means_comparison(df, dimensions, xlabel=None, ylabel=None,
                                                  levels=[-0.50, -0.32, -0.17, -0.05, 0.05, 0.17, 0.32, 0.50],
                                                  data_standardized=False, output_path=None, savefig_kws=None):
    """
    Plots the normalized curve used for computing the optimal number of clusters.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the variables used for clustering.
    dimensions : list
        List of variables of interest.
        *Note these must be internal variables, ie, variables used for clustering*
    xlabel : str, default=None
        x-label name/description.
    ylabel : str, default=None
        y-label name/description.
    levels : list or `numpy.array`
        Values to be used as cuts for color intensity.
        Default values: [-0.50, -0.32, -0.17, -0.05, 0.05, 0.17, 0.32, 0.50]
    data_standardized : bool, default=False
        If data are standardized, comparison to global mean is based solely on the mean per cluster.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    df_diff = compare_cluster_means_to_global_means(df, dimensions, data_standardized=data_standardized)
    colors = sns.color_palette("BrBG", n_colors=9)
    cmap, norm = matplotlib.colors.from_levels_and_colors(levels, colors, extend="both")
    width = min(len(dimensions), 20)
    height = min(df['cluster'].nunique(), 8)
    fig, ax = plt.subplots(figsize=(width, height))
    im = ax.imshow(df_diff[dimensions].values, cmap=cmap, norm=norm)
    ax.set(xticks=range(len(dimensions)), yticks=range(df_diff.shape[0]),
           xticklabels=list(map(str.upper, dimensions)), yticklabels=df_diff['cluster'])
    ax.tick_params(axis='x', rotation=40, labelsize=10)
    ax.tick_params(axis='y', labelsize=10)
    ax.set_xlabel('' if xlabel is None else xlabel, fontsize=12, weight='bold', labelpad=15)
    ax.set_ylabel('' if ylabel is None else ylabel, fontsize=12, weight='bold', labelpad=15)
    for i in range(len(df_diff['cluster'].unique())):
        for j in range(len(dimensions)):
            val = df_diff.loc[i, dimensions[j]]
            val_str = '{:.2f}'.format(val)
            if val < 0:
                val_str = '- ' + '{:.2f}'.format(-val)

            text = ax.text(j, i, val_str,
                           ha="center", va="center", color="black", fontsize=11, fontweight='ultralight',
                           fontstretch='ultra-expanded')

    # Turns off grid on the left axis
    ax.grid(False)
    fig.colorbar(im, ax=ax)
    fig.tight_layout()
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_distribution_by_cluster(df, cluster_labels, xlabel=None, ylabel=None, sharex=True, sharey=True,
                                 output_path=None, savefig_kws=None):
    """
    Plots the violin plots per cluster and *continuous* variables of interest to understand differences in their
    distributions by cluster.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the variables used for clustering.
    cluster_labels : `numpy.array` or list
        Array with cluster labels.
        *Note this array should have the same length as df and observations be in the same order*.
    xlabel : str, default=None
        x-label name/description.
    ylabel : str, default=None
        y-label name/description.
    sharex : bool, default=True
        If True, all subplots share the x-axis.
    sharey : bool, default=True
        If True, all subplots share the y-axis.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    nclusters = len(np.unique(cluster_labels))
    nvars = df.shape[1]
    ncols = max(1, min(nvars, 18//nclusters))
    if ncols > 3 and nvars % ncols > 0:
        if nvars % 3 == 0:
            ncols = 3
        elif nvars % 2 == 0:
            ncols=2

    nrows = nvars // ncols + (nvars % ncols > 0)
    fig, axs = plt.subplots(nrows, ncols, figsize=(max(nclusters * ncols, 9), 5 * nrows), sharex=sharex, sharey=sharey)

    i = 0
    for col in df.columns:
        ax = get_axis(i, axs, ncols, nrows)
        sns.violinplot(y=df[col], x=cluster_labels, linewidth=1, ax=ax)
        plt.setp(ax.collections, alpha=.4)
        sns.boxplot(y=df[col], x=cluster_labels, width=0.2, linewidth=1, color='grey', ax=ax)
        sns.stripplot(y=df[col], x=cluster_labels, alpha=0.7, size=3, ax=ax)
        ax.set_ylabel(col if ylabel is None else ylabel, fontsize=12, labelpad=15)
        if i // ncols == nrows-1:
            ax.set_xlabel('cluster' if xlabel is None else xlabel, fontsize=12, labelpad=15)
        i += 1

    while i < ncols * nrows:
        ax = get_axis(i, axs, ncols, nrows)
        ax.axis('off')
        i += 1

    fig.tight_layout(pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_clusters_2D(x, y, hue, df, style_kwargs=dict(), output_path=None, savefig_kws=None):
    """
    Plots two 2D plots:
     - A scatter plot styled by the categorical variable `hue`.
     - A 2D plot comparing cluster centroids and optionally the density area.

    Parameters
    ----------
    x : `numpy.array` or list
        x-coordinate data.
    y : `numpy.array` or list
        y-coordinate data.
    hue : `numpy.array` or list
        Array with categorical values to be used for color styling.
    df : `pandas.DataFrame`
        DataFrame containing the data.
    style_kwargs : dict, default=empty dict
        Dictionary with optional styling parameters.
        List of parameters:
         - palette : matplotlib palette to be used. default='gnuplot'
         - vline_color : color to be used for vertical line (used for plotting x mean value). default='#11A579'
         - hline_color : color to be used for horizontal line (used for plotting y mean value). default='#332288'
         - kdeplot : boolean to display density area of points (using seabonr.kdeplot). default=True
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    # Style params
    palette = 'gnuplot'
    if style_kwargs.get('palette'):
        palette = style_kwargs.get('palette')

    vline_color = '#11A579'
    if style_kwargs.get('vline_color'):
        vline_color = style_kwargs.get('vline_color')

    hline_color = '#332288'
    if style_kwargs.get('hline_color'):
        vline_color = style_kwargs.get('hline_color')

    kdeplot = True
    if style_kwargs.get('kdeplot') is not None:
        kdeplot = style_kwargs.get('kdeplot')

    fig, axs = plt.subplots(1, 2, figsize=(14, 5), sharey=True, sharex=True)

    x_range = df[x].max() - df[x].min()
    xmin = df[x].min() - x_range * 0.05
    xmax = df[x].max() + x_range * 0.05
    y_range = df[y].max() - df[y].min()
    ymin = df[y].min() - y_range * 0.05
    ymax = df[y].max() + y_range * 0.05

    # Left-hand side plot: Scatter plot colored by cluster category
    sns.scatterplot(x=x, y=y, hue=hue, data=df.sort_values(hue), alpha=0.3, palette=palette, linewidth=0, ax=axs[0])
    axs[0].vlines(df[x].mean(), ymin=ymin, ymax=ymax, color=vline_color, linewidth=1.15, linestyles='--', label=f'Mean {x}')
    axs[0].hlines(df[y].mean(), xmin=xmin, xmax=xmax, color=hline_color, linewidth=1.15, linestyles='--',
                  label=f'Mean {y}')
    axs[0].set_xlabel(x, fontsize=12)
    axs[0].set_ylabel(y, fontsize=12)
    axs[0].set_title('Scatter plot by cluster', fontsize=13)
    axs[0].set_xlim(xmin, xmax)
    axs[0].set_ylim(ymin, ymax)

    # Right-hand side plot: Cluster centroids with optional kernel density area
    sns.scatterplot(x=x, y=y, hue=hue, data=df.groupby(hue).mean().reset_index(),
                    alpha=1, palette=palette, linewidth=0, marker='X', s=100, ax=axs[1])

    if kdeplot:
        sns.kdeplot(x=x, y=y, hue=hue, data=df.sort_values(hue), levels=1, alpha=0.2, palette=palette,
                    ax=axs[1])

    axs[1].vlines(df[x].mean(), ymin=ymin, ymax=ymax, color=vline_color, linewidth=1, linestyles='--',
                  label=f'Mean {x}')
    axs[1].hlines(df[y].mean(), xmin=xmin, xmax=xmax, color=hline_color, linewidth=1, linestyles='--',
                  label=f'Mean {y}')
    axs[1].set_xlabel(x, fontsize=12)
    axs[1].set_ylabel(y, fontsize=12)
    axs[1].set_title('Cluster centroids', fontsize=13)

    axs[0].legend(fontsize=11, title='', title_fontsize=12, labelspacing=0.5,
                  loc=(0.93, 0.5 - 0.167 * (df[hue].nunique() // 4)))
    axs[1].legend(fontsize=11, title='', title_fontsize=12, labelspacing=0.5,
                  loc=(0.93, 0.5 - 0.167 * (df[hue].nunique() // 4)))

    fig.tight_layout(pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_cat_distribution_by_cluster(ct, cat_label=None, cluster_label=None, output_path=None, savefig_kws=None):
    """
    Plots the relative contingency table of the clusters with a categorical variable as a stacked bar plot.

    Parameters
    ----------
    ct : `pandas.DataFrame`
        DataFrame with the relative contingency table.
    cat_label : str, default=None
        Name/Description of the categorical variable to be displayed.
    cluster_label : str, default=None
        Name/Description of the cluster variable to be displayed.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    plt.figure(figsize=(11, 0.625 * len(ct.index)))
    colors = sns.color_palette("YlGnBu", n_colors=len(ct.columns))
    left = np.array([0] * len(ct.index))

    i = 0
    for col in ct.columns:
        widths = ct[col].values
        plt.barh(ct.index, widths, left=left, label=col, color=colors[i], height=0.7)

        xcenters = left + widths / 2
        for y, (x, w) in enumerate(zip(xcenters, widths)):
            if w > 0.05:
                color = '#737373' if i < len(ct.columns)//2 else '#d9d9d9'
                plt.text(x, y, f'{str(np.round(w * 100, 1))}%', ha='center', va='center', color=color, fontsize=12,
                         weight='light')

        left = left + ct[col].values
        i += 1

    ncol = 5
    if len(ct.columns) % ncol > 0:
        if len(ct.columns) % 6 == 0:
            ncol = 6
        elif len(ct.columns) % 4 == 0:
            ncol = 4

    plt.gca().invert_yaxis()
    plt.legend(ncol=ncol, loc='lower center', bbox_to_anchor=(0.5, 1), fontsize=12, title=cat_label,
               title_fontsize=13)
    plt.ylabel(cluster_label, fontsize=12, weight='bold', labelpad=15)
    plt.yticks(ticks=range(len(ct.index)), labels=list(ct.index), fontsize=11)
    plt.xticks([])
    plt.xlim(0, 1)

    plt.tight_layout()
    savefig(output_path=output_path, savefig_kws=savefig_kws)



"""Utils for visualization"""
# Author: Miguel Alvarez-Garcia

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import seaborn as sns

from ..utils import savefig


def missing_values_heatmap(df, output_path=None, savefig_kws=None):
    """
    Plots a heatmap to visualize missing values (light color).

    Parameters
    ----------
    df : `pandas.DataFrame`
       DataFrame containing the data.
    output_path : str, default=None
       Path to save figure as image.
    savefig_kws : dict, default=None
       Save figure options.
    """
    fig, ax = plt.subplots(figsize=(18, 10))
    sns.heatmap(df.isnull().astype(int), cbar=False)
    fig.tight_layout()
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_imputation_pairs_scatter(df, imputation_pairs, sample_frac=1.0, scatter_kws=None, line_kws=None,
                                  output_path=None, savefig_kws=None):
    """
    Plots a grid of scatter plots with every pair of independent-dependent variables used for one-to-one model-based
    imputation.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    imputation_pairs : `pandas.DataFrame`
        Imputation pairs as returned by the function `ldbx.data_processing.imputation_pairs()`.
    sample_frac : float, default=1.0
        If < 1 a random sample of every pair of variables will be plotted.
    {scatter,line}_kws : dict, default=None
        Additional keyword arguments to pass to the scatter and line plots.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    scatter_kws = scatter_kws if scatter_kws else dict(color='blue', alpha=0.05)
    line_kws = line_kws if line_kws else dict(color='red')

    ncols = min(4, imputation_pairs.shape[0])
    nrows = int(np.ceil(imputation_pairs.shape[0] / ncols))

    fig, axs = plt.subplots(nrows, ncols, figsize=(4 * ncols, 4 * nrows))
    df_sample = df.sample(frac=sample_frac, random_state=42)
    i = 0
    for idx, row in imputation_pairs.iterrows():
        ax = axs[i // ncols, i % ncols]
        sns.regplot(x=row['var2'], y=row['var1'], data=df_sample, scatter_kws=scatter_kws, line_kws=line_kws, ax=ax)
        ax.set_xlabel(f"{row['var2']} ({'{:.2f}'.format(row['missing_var2'])}% NA)")
        ax.set_ylabel(f"{row['var1']} ({'{:.2f}'.format(row['missing_var1'])}% NA)")
        i += 1

    fig.tight_layout(pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_imputation_distribution_assessment(df_prior, df_posterior, imputed_vars, sample_frac=1.0, prior_kws=None,
                                            posterior_kws=None, output_path=None, savefig_kws=None):
    """
    Plots a distribution comparison of each variable with imputed variables, before and after imputation.

    Parameters
    ----------
    df_prior : `pandas.DataFrame`
        DataFrame containing the data before imputation.
    df_posterior : `pandas.DataFrame`
        DataFrame containing the data after imputation.
    imputed_vars : list
        List of variables with imputed variables.
    sample_frac : float, default=1.0
        If < 1 a random sample of every pair of variables will be plotted.
    {prior,posterior}_kws : dict, default=None
        Additional keyword arguments to pass to the kdeplot (https://seaborn.pydata.org/generated/seaborn.kdeplot.html).
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    prior_kws = prior_kws if prior_kws else dict(color='#7F3C8D')
    posterior_kws = posterior_kws if posterior_kws else dict(color='#11A579')

    ncols = min(4, len(imputed_vars))
    nrows = int(np.ceil(len(imputed_vars) / ncols))

    fig, axs = plt.subplots(nrows, ncols, figsize=(4 * ncols, 4 * nrows))
    df_prior_sample = df_prior.sample(frac=sample_frac, random_state=42)
    i = 0
    for ivar in imputed_vars:
        ax = axs[i // ncols, i % ncols]
        sns.kdeplot(ivar, data=df_prior_sample, label='Before imputation', ax=ax, **prior_kws)
        sns.kdeplot(ivar, data=df_posterior.loc[df_prior_sample.index], label='After imputation', ax=ax,
                    **posterior_kws)
        ax.legend()
        i += 1

    fig.tight_layout(pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_variable_graph_partitioning_components(edges, connected_components, graph_style_kws=None, output_path=None,
                                                savefig_kws=None):
    """
    Plots a connected components of a graph.
    **Note** this function relies on `ldbx.data_processing.variable_graph_partitioning()` for computing the edges and
    connected components.

    Parameters
    ----------
    edges : list of tuples
        List of graph edges
    connected_components : list of sets
        List of connected components.
    graph_style_kws : dict, default=None
        Additional keyword arguments to style graph plots
        (see https://networkx.org/documentation/stable/reference/drawing.html)
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    if graph_style_kws is None:
        graph_style_kws = dict(node_size=16, width=0.4, edge_color='grey', node_color='red', with_labels=True)

    nrows = np.shape(connected_components)[0]
    fig, axs = plt.subplots(nrows, 1, figsize=(12, 5*nrows))

    for i in range(nrows):
        fedges = []
        for edge in edges:
            if edge[0] in connected_components[i]:
                fedges.append(edge)

        G = nx.Graph()
        G.add_nodes_from(connected_components[i])
        G.add_edges_from(fedges)

        nx.draw(G, ax=axs[i], **graph_style_kws)
        savefig(output_path=output_path, savefig_kws=savefig_kws)



"""Table statistics utils for clustering"""
# Author: Miguel Alvarez-Garcia


def compare_cluster_means_to_global_means(df, dimensions, data_standardized=False, output_path=None):
    df_agg = df.groupby('cluster')[dimensions].mean()
    df_agg_diff = df_agg.copy()
    if data_standardized:
        df_agg_diff = df_agg
    else:
        mean_array = df[dimensions].mean().values
        for idx, row in df_agg.iterrows():
            df_agg_diff.loc[idx, dimensions] = (row[dimensions] - mean_array) / mean_array

    df_agg_diff = df_agg_diff.reset_index()
    if output_path is not None:
        df_agg_diff.to_csv(output_path, index=False)

    return df_agg_diff



"""Clustering with KMeans++ as default algorithm"""
# Author: Miguel Alvarez-Garcia

import pandas as pd
import statsmodels.api as sm

from scipy.stats import chi2_contingency
from sklearn.cluster import AgglomerativeClustering, KMeans
from sklearn.metrics import calinski_harabasz_score, davies_bouldin_score, silhouette_score
from sklearn.preprocessing import MinMaxScaler
from statsmodels.formula.api import ols
from .table_utils import *
from .utils import *
from .viz_utils import *

__metrics__ = ['inertia', 'davies_bouldin_score', 'silhouette_score', 'calinski_harabasz_score']
METRIC_NAMES = dict(zip(__metrics__,
                        ['Weighted Sum of Squared Distances', 'Davies Bouldin Score', 'Silhouette Score',
                         'Calinski and Harabasz score']))
KMEANS = ['kmeans', 'kmeans++']
HIERARCHICAL_WARD = ['ward', 'hierarchical', 'agglomerative']


class Clustering:
    """
    Clustering class

    Parameters
    ----------
    df : `pandas:DatasFrame`
        DataFrame with main data
    algorithms : str or list, default='kmeans'
        Algorithm/s to be used for clustering.
        By default, [K-Means++](https://scikit-learn.org/stable/modules/clustering.html#k-means)
    normalize : bool, default=True
        Whether to apply data normalization for fair comparisons between variables. By default it is applied. In case
        dimensionality reduction is applied beforehand, normalization should not be applied.
    """
    def __init__(self,
                 df,
                 algorithms='kmeans',
                 normalize=True):

        # Normalize variables for fair comparisons
        self.normalize = normalize
        self.df = df.copy()
        if normalize:
            mms = MinMaxScaler()
            self.df = pd.DataFrame(mms.fit_transform(df), columns=df.columns)

        self.dimensions_ = list(df.columns)

        if not isinstance(algorithms, list):
            algorithms = [algorithms]
        self.algorithms = list(map(str.lower, algorithms))

        self.instances_ = dict()
        for algorithm in self.algorithms:
            if algorithm in KMEANS:
                self.instances_[algorithm] = KMeans(random_state=42)
            elif algorithm in HIERARCHICAL_WARD:
                self.instances_[algorithm] = AgglomerativeClustering()
            else:
                raise RuntimeWarning(f'''Algorithm {algorithm} is not supported.
                                     Supported algorithms are: KMeans and Hierarchical clustering''')

        self.scores_ = dict()
        self._initialize_scores()

        self.metric_ = 'inertia'
        self.optimal_config_ = None

    def _initialize_scores(self):
        for algorithm in self.algorithms:
            if algorithm in KMEANS + HIERARCHICAL_WARD:
                self.scores_[algorithm] = []

    def _compute_clusters(self, algorithm, n_clusters):
        self.instances_[algorithm].set_params(n_clusters=n_clusters)
        self.instances_[algorithm].fit(self.df[self.dimensions_])
        self.labels_ = self.instances_[algorithm].labels_

    def _compute_optimal_clustering_config(self, metric, cluster_range, weights):
        optimal_list = []
        for algorithm in self.algorithms:
            for nc in range(*cluster_range):
                self.instances_[algorithm].set_params(n_clusters=nc)
                self.instances_[algorithm].fit(self.df[self.dimensions_])

                if metric == 'inertia':
                    self.scores_[algorithm].append(
                        weighted_sum_of_squared_distances(self.df[self.dimensions_], self.instances_[algorithm].labels_,
                                                          weights))
                elif metric == 'davies_bouldin_score':
                    self.scores_[algorithm].append(
                        1 if nc == 1 else davies_bouldin_score(self.df[self.dimensions_],
                                                               self.instances_[algorithm].labels_))
                elif metric == 'silhouette_score':
                    self.scores_[algorithm].append(
                        0 if nc == 1 else silhouette_score(self.df[self.dimensions_],
                                                           self.instances_[algorithm].labels_))

                elif metric == 'calinski_harabasz_score':
                    self.scores_[algorithm].append(
                        1 if nc == 1 else calinski_harabasz_score(self.df[self.dimensions_],
                                                                  self.instances_[algorithm].labels_))

            if len(range(*cluster_range)) > 1:
                kl = KneeLocator(x=range(*cluster_range), y=self.scores_[algorithm], curve='convex',
                                 direction='decreasing')
                optimal_list.append((algorithm, kl.knee, self.scores_[algorithm][kl.knee-cluster_range[0]]))
            else:
                optimal_list.append((algorithm, cluster_range[0], self.scores_[algorithm][0]))

        return min(optimal_list, key=lambda t: t[2])

    def compute_clusters(self, n_clusters=None, metric='inertia', max_clusters=10, prefix=None, weights=None):
        """
        Calculates clusters.
        If more than one algorithm is passed in the class constructor, first, the optimal number of clusters
        is computed for each algorithm based on the metric passed to the method. Secondly, the algorithm that
        provides the best performance for the corresponding optimal number of clusters is selected.
        Therefore, the result shows the clusters calculated with the best performing algorithm based on the
        criteria explained above.

        Parameters
        ----------
        n_clusters : int, default=None
            Number of clusters to be computed. If n_clusers=None, the optimal number of clusters is computed
            using the elbow/knee method.
            For optimal number of cluster calculation, the Python package [kneed](https://pypi.org/project/kneed/)
            is used, where the method presented in [this paper](https://raghavan.usc.edu//papers/kneedle-simplex11.pdf)
            is implemented.
        metric : str, default='inertia'
            Metric to use in order to compare different algorithms and, if applicable,to calculate the optimal number
            of clusters.
            By default, inertia is used. Supported metrics: ['inertia', 'davies_bouldin_score', 'silhouette_score',
            'calinski_harabasz_score']
        max_clusters : int, default=10
            In case of optimal search, this parameter limits the maximum number of clusters allowed.
        prefix : str, default=None
            Used for cluster naming. Naming format: `f'{prefix}_{x}'`
        weights : `numpy.array`, default=None
            In case observations have different weights.
            *Note this is not implemented yet.*

        Returns
        ----------
        labels_ : `numpy.array`
            Cluster label of each observation.
        """
        if metric not in __metrics__:
            raise RuntimeError(f'''Metric {metric} not supported.
                               Supported metrics: {__metrics__}''')

        self._initialize_scores()
        self.metric_ = metric

        # Compute optimal number of clusters
        cluster_range = []
        if n_clusters is not None:
            cluster_range = [n_clusters, n_clusters + 1]
        else:
            cluster_range = [1, max_clusters + 1]

        self.optimal_config_ = self._compute_optimal_clustering_config(metric, cluster_range, weights)

        if self.optimal_config_ is None:
            raise RuntimeError('Optimal cluster configuration not available')

        self._compute_clusters(self.optimal_config_[0], self.optimal_config_[1])

        # Might be interesting to only keep one
        self.df['cluster'] = self.labels_
        self.df['cluster_cat'] = self.labels_
        if prefix is not None:
            name_len = len(str(max(self.labels_)))
            self.df['cluster_cat'] = list(map(lambda x: f'{prefix}_{str.zfill(str(x), name_len)}', self.labels_))

        return self.labels_

    def describe_clusters(self, df_ext=None, variables=None, cluster_filter=None, statistics=['mean', 'median', 'std'],
                          output_path=None):
        """
        Describes clusters based on internal or external *continuous* variables.
        For categorical variables use `describe_clusters_cat()`.

        Parameters
        ----------
        df_ext : `pandas.DataFrame`, default=None
            Optional. For describing clusters based on external variables.
            This DataFrame must only contain the variables of interest. The order of the observations must be the same
            as that of the base DataFrame.
        variables : str or list, default=None
            List of variables (internal or external) for describing clusters. This parameter is optional and should
            be used when only a subset of the variable is of interest.
        cluster_filter : str or list, default=None
            In case the descriptive statistics of interest only applies to a subset of the calculated clusters.
        statistics : str or list, default=['mean', 'median', 'std']
            Statistics to use for describing clusters.
            *Note any statistics supported by Pandas can be used. This includes the `describe` function*
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        res : `pandas.DataFrame`
            DataFrame with the cluster description.
        """
        if df_ext is not None:
            df_ext = df_ext.copy()
            df_ext['cluster'] = self.labels_
        else:
            df_ext = self.df[self.dimensions_ + ['cluster']]

        if variables is not None:
            if not isinstance(variables, list):
                variables = [variables]
            df_ext = df_ext[variables + ['cluster']]
        else:
            variables = df_ext.columns.to_list()
            variables.remove('cluster')

        if cluster_filter is None:
            cluster_filter = list(df_ext['cluster'].unique())
        if not isinstance(cluster_filter, list):
            cluster_filter = [cluster_filter]

        res = df_ext[df_ext['cluster'].isin(cluster_filter)].groupby('cluster').agg(
            dict(zip(list(variables), [statistics] * len(variables)))).reset_index()

        if output_path is not None:
            res.to_csv(output_path, index=False)

        return res

    def describe_clusters_cat(self, cat_array, cat_name=None, order=None, normalize=False, output_path=None):
        """
        Describes clusters based on  external *categorical* variables. The result is a contingency table.
        For continuous variables use `describe_clusters()`.

        Parameters
        ----------
        cat_array : `pandas.Series` or `numpy.array`
            Values of categorical variable.
            The order of the observations must be the same as that of the base DataFrame.
        cat_name : str, default=None
            Name of the categorical variable.
        order : list or `numpy.array`, default=None
            In case categories should be displayed in a specific order.
        normalize : boolean, default=False
            If True, results are row-normalized.
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        freq : `pandas.DataFrame`
            DataFrame with a cluster description based on the passed categorical variable.
        """
        freq = pd.crosstab(index=self.df['cluster_cat'], columns=cat_array, rownames=['Clusters'], colnames=[cat_name])
        if order is not None:
            freq = freq[order]

        if normalize:
            freq['total'] = freq.sum(1)
            for col in freq.columns[:-1]:
                freq[col] = freq[col] / freq['total']
            freq = freq.drop(columns='total')

        if output_path is not None:
            freq.to_csv(output_path, index=False)

        return freq

    def compare_cluster_means_to_global_means(self, df_original=None, output_path=None):
        """
        For every cluster and every internal variable, the relative difference between the intra-cluster mean
        and the global mean.

        Parameters
        ----------
        df_original : `pandas.DataFrame`, default=None
            In case the comparison wants to be made with the original variables and values. Note it is assumed that both
            the original dataframe and the one used for clustering have observations is the same order.
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        df_agg_diff : `pandas.DataFrame`
            DataFrame with the comparison.
        """
        if df_original is not None:
            var_names = list(df_original.columns)
            df_original = df_original.copy()
            df_original['cluster'] = self.labels_
            return compare_cluster_means_to_global_means(df_original, var_names, output_path=output_path)
        else:
            return compare_cluster_means_to_global_means(self.df, self.dimensions_, data_standardized=not self.normalize,
                                                         output_path=output_path)

    def anova_tests(self, df_test=None, vars_test=None, cluster_filter=None, output_path=None):
        """
        Runs ANOVA tests for a given set of continuous variables (internal or external) to test dependency with
        clusters.

        Parameters
        ----------
        df_test : `pandas.DataFrame`, default=None
            Optional. For running tests with external continuous variables.
            This DataFrame must only contain the variables of interest. The order of the observations must be the same
            as that of the base DataFrame.
        vars_test : str, default=None
            List of variables (internal or external) tu run tests on. This parameter is optional and should
            be used when only a subset of the variable is of interest.
        cluster_filter : str or list, default=None
            In case the tests should only be applied on a subset of  clusers.
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        res : `pandas.DataFrame`
            DataFrame with the corresponding test statistics.
        """
        if df_test is not None:
            df_test = df_test.copy()
            df_test['cluster'] = self.labels_
        else:
            df_test = self.df[self.dimensions_ + ['cluster']]

        if vars_test is not None:
            if not isinstance(vars_test, list):
                vars_test = [vars_test]
            df_test = df_test[vars_test + ['cluster']]

        if cluster_filter is None:
            cluster_filter = list(df_test['cluster'].unique())
        if not isinstance(cluster_filter, list):
            cluster_filter = [cluster_filter]

        res = []
        col_names = []
        variables = df_test.columns.to_list()
        variables.remove('cluster')
        for var in variables:
            model = ols(f'{var} ~ C(cluster)', data=df_test[df_test['cluster'].isin(cluster_filter)]).fit()
            aov_table = sm.stats.anova_lm(model, typ=1)
            res.append([var] + aov_table.iloc[0].to_list())
            if len(col_names) == 0:
                col_names = ['var_name'] + aov_table.columns.tolist()

        res = pd.DataFrame(res, columns=col_names)
        if output_path is not None:
            res.to_csv(output_path, index=False)

        return res

    def chi2_test(self, cat_array):
        """
        Runs Chi-squared tests for a given categorical variable to test dependency with clusters.

        Parameters
        ----------
        cat_array : `pandas.Series` or `numpy.array`
            Values of categorical variable.
            The order of the observations must be the same as that of the base DataFrame.

        Returns
        ----------
        dict : dictionary
            Dictionary with the corresponding test statistics.
        """
        contingency_t = self.describe_clusters_cat(cat_array)
        test_res = chi2_contingency(contingency_t.values)
        return dict(zip(['chi2', 'p', 'dof'], test_res[:-1]))

    def plot_score_comparison(self, output_path=None, savefig_kws=None):
        """
        Plots the comparison in performance between the different clustering algorithms.

        Parameters
        ----------
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        metric_name = METRIC_NAMES[self.metric_]

        cluster_range = [1, self.optimal_config_[1] + 1]
        if len(self.scores_[self.optimal_config_[0]]) > 1:
            cluster_range = [1, len(self.scores_[self.optimal_config_[0]]) + 1]
        else:
            cluster_range = [self.optimal_config_[1], self.optimal_config_[1] + 1]

        plot_score_comparison(self.scores_, cluster_range, metric_name, output_path, savefig_kws)

    def plot_optimal_components_normalized(self, output_path=None, savefig_kws=None):
        """
        Plots the normalized curve used for computing the optimal number of clusters.

        Parameters
        ----------
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        if len(self.scores_[self.optimal_config_[0]]) > 1:
            plot_optimal_components_normalized(self.scores_[self.optimal_config_[0]],
                                               len(self.scores_[self.optimal_config_[0]]),
                                               METRIC_NAMES[self.metric_],
                                               output_path, savefig_kws)
        else:
            raise RuntimeError('This plot can only be used when `cluster_range` contains at least 2 values')

    def plot_clustercount(self, output_path=None, savefig_kws=None):
        """
        Plots a bar plot with cluster counts.

        Parameters
        ----------
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        plot_clustercount(self.df, output_path, savefig_kws)

    def plot_cluster_means_to_global_means_comparison(self, df_original=None, xlabel=None, ylabel=None,
                                                      levels=[-0.50, -0.32, -0.17, -0.05, 0.05, 0.17, 0.32, 0.50],
                                                      output_path=None, savefig_kws=None):
        """
        Plots the normalized curve used for computing the optimal number of clusters.

        Parameters
        ----------
        df_original : `pandas.DataFrame`, default=None
            In case the comparison wants to be made with the original variables and values. Note it is assumed that both
            the original dataframe and the one used for clustering have observations is the same order.
        xlabel : str, default=None
            x-label name/description.
        ylabel : str, default=None
            y-label name/description.
        levels : list or `numpy.array`
            Values to be used as cuts for color intensity.
            Default values: [-0.50, -0.32, -0.17, -0.05, 0.05, 0.17, 0.32, 0.50]
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """

        if df_original is not None:
            var_names = list(df_original.columns)
            df_original = df_original.copy()
            df_original['cluster'] = self.labels_

            plot_cluster_means_to_global_means_comparison(df_original, var_names, xlabel, ylabel, levels,
                                                          data_standardized=False, output_path=output_path,
                                                          savefig_kws=savefig_kws)
        else:
            plot_cluster_means_to_global_means_comparison(self.df, self.dimensions_, xlabel, ylabel, levels,
                                                          data_standardized=not self.normalize, output_path=output_path,
                                                          savefig_kws=savefig_kws)

    def plot_distribution_comparison_by_cluster(self, df_ext=None, xlabel=None, ylabel=None, output_path=None,
                                                savefig_kws=None):
        """
        Plots the violin plots per cluster and *continuous* variables of interest to understand differences in their
        distributions by cluster.

        Parameters
        ----------
        df_ext : `pandas.DataFrame`, default=None
            DataFrame containing external variables for comparison.
            If None, internal variables will be compared.
        xlabel : str, default=None
            x-label name/description.
        ylabel : str, default=None
            y-label name/description.
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        sharey = self.normalize
        if df_ext is None:
            df_ext = self.df[self.dimensions_]
        else:
            sharey = False

        plot_distribution_by_cluster(df_ext, self.labels_, xlabel=xlabel, ylabel=ylabel, sharey=sharey,
                                     output_path=output_path, savefig_kws=savefig_kws)

    def plot_clusters_2D(self, coor1, coor2, style_kwargs=dict(), output_path=None, savefig_kws=None):
        """
        Plots two 2D plots:
         - A scatter plot styled by the categorical variable `hue`.
         - A 2D plot comparing cluster centroids and optionally the density area.

        Parameters
        ----------
        coor1 : int, str, or `pandas.Series`
            If int, it represents the id of the internal variable to be used.
            If str, it must be an internal variable name.
            If `pandas.Series`, it is assumed it is an external variable.
        coor2 : int, str, or `pandas.Series`
            If int, it represents the id of the internal variable to be used.
            If str, it must be an internal variable name.
            If `pandas.Series`, it is assumed it is an external variable.
        style_kwargs : dict, default=empty dict
            Dictionary with optional styling parameters.
            List of parameters:
             - palette : matplotlib palette to be used. default='gnuplot'
             - vline_color : color to be used for vertical line (used for plotting x mean value). default='#11A579'
             - hline_color : color to be used for horizontal line (used for plotting y mean value). default='#332288'
             - kdeplot : boolean to display density area of points (using seabonr.kdeplot). default=True
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        if isinstance(coor1, int):
            coor1 = self.dimensions_[coor1]
        if isinstance(coor2, int):
            coor2 = self.dimensions_[coor2]

        df_ext = self.df
        if isinstance(coor1, pd.Series):
            if not isinstance(coor2, pd.Series):
                raise TypeError('For external variables, both coor1 and coor2 need to be of type pandas.Series')
            df_ext = pd.DataFrame(data={coor1.name: coor1,
                                        coor2.name: coor2,
                                        'cluster_cat': self.df['cluster_cat'].values})
            coor1 = coor1.name
            coor2 = coor2.name

        hue = 'cluster_cat'
        plot_clusters_2D(coor1, coor2, hue, df_ext, style_kwargs=style_kwargs, output_path=output_path,
                         savefig_kws=savefig_kws)

    def plot_cat_distribution_by_cluster(self, cat_array, cat_label=None, cluster_label=None, output_path=None,
                                         savefig_kws=None):
        """
        Plots the relative contingency table of the clusters with a categorical variable as a stacked bar plot.

        Parameters
        ----------
        cat_array : `numpy.array` or list
            Array with categorical values.
            *Note its length must be the same as self.df and observations be in the same order*.
        cat_label : str, default=None
            Name/Description of the categorical variable to be displayed.
        cluster_label : str, default=None
            Name/Description of the cluster variable to be displayed.
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        ct = self.describe_clusters_cat(cat_array, normalize=True)
        plot_cat_distribution_by_cluster(ct, cat_label, cluster_label, output_path, savefig_kws)



"""Utils for the classification module"""
# Author: Miguel Alvarez-Garcia

import logging
import numpy as np
import pandas as pd
import shap

from sklearn.feature_selection import RFECV
from sklearn.model_selection import GridSearchCV
from ..utils import compute_high_corr_pairs, compute_highly_related_categorical_vars, compute_highly_related_mixed_vars

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(name)s: %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger(__name__)


def shap_importances(classifier, X):
    """
    Computes shap importance values as the combined average of the absolute values of the shap values
    for all classes.

    Parameters
    ----------
    classifier : object
        Classification model supported by SHAP (e.g., sklearn models or xgboost).
    X : `numpy.ndarray` or `pandas.DataFrame`
        Observations.

    Returns
    ----------
    importances : `pandas.DataFrame`
        DataFrame with predictors and their corresponding importance.
    """
    explainer = shap.Explainer(classifier)
    shap_values = explainer(X)
    shap_importance_values = [np.abs(shap_values.values[:, i, :]).mean() for i in range(shap_values.values.shape[1])]
    importances = pd.DataFrame(
        data={'variable_name': shap_values.feature_names, 'shap_importance': shap_importance_values}).sort_values(
        'shap_importance', ascending=False).reset_index(drop=True)
    return importances


def _compute_highly_related_pairs(df, num_vars=None, cat_vars=None, num_kws=None, mixed_kws=None, cat_kws=None):
    """
    Computes strongly related pairs of variables. Depending on the type of variables, a correlation coefficient
    (numerical variables), partial eta squared (mixed-type variables), or mutual information (categorical variables) is
    used to measured the strength of the relationship of each pair.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    num_vars : str, list, series, or vector array
        Numerical variable name(s).
    cat_vars : str, list, series, or vector array
        Categorical variable name(s).
    {num,mixed,cat}_kws : dict, default=None
        Additional keyword arguments to pass to `compute_high_corr_pairs()`, `compute_highly_related_mixed_vars()`, and
        `compute_highly_related_categorical_vars()`.

    Returns
    ----------
    final_pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables (var1: variable with values to impute; var2: variable to be
        used as independent variable for model-based imputation).
    """
    if num_vars is None and cat_vars is None:
        raise ValueError('Numerical or categorical variable lists are required.')

    # Numerical variable pairs (correlation)
    num_pairs = pd.DataFrame()
    if num_vars:
        num_kws = num_kws if num_kws else dict()
        num_pairs = compute_high_corr_pairs(df[num_vars], **num_kws)

    # Mixed variable pairs (cross correlation)
    mixed_pairs = pd.DataFrame()
    if num_vars and cat_vars:
        mixed_kws = mixed_kws if mixed_kws else dict()
        mixed_pairs = compute_highly_related_mixed_vars(df, num_vars, cat_vars, **mixed_kws)

    # Categorical variable pairs (mutual information)
    cat_pairs = pd.DataFrame()
    if cat_vars:
        cat_kws = cat_kws if cat_kws else dict()
        cat_pairs = compute_highly_related_categorical_vars(df[cat_vars], **cat_kws)

    final_pairs = pd.concat([num_pairs, mixed_pairs, cat_pairs], ignore_index=True)
    return final_pairs


def feature_selection(df, original_features, target, classifier, num_vars=None, cat_vars=None, features_to_keep=[],
                      num_kws=None, mixed_kws=None, cat_kws=None, rfecv_kws=None):
    """
    Performs feature selection in three steps:
        - First, if some features must be kept (informed in `features_to_keep`), those other
          features that highly related with those in `features_to_keep` are removed.

        - Next, a classifier is iteratively trained to obtain feature shap importances.
          In each iteration, the feature with the highest shape importance which has not been previously
          visited is selected and all other highly related features are removed.

        - Finally, Recursive Feature Elimination with Cross-Validation (RFECV) is applied on the remaining
          features.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing, at least, all variables in `original_features`
    original_features : `numpy.array` or list
        Array of original features to select from. `original_features` = `num_vars` + `cat_vars`.
    target : `numpy.array` or list
        Array with target values. This is the dependent variable values.
    classifier : estimator
        Should be an instance of `scikit-learn.Estimator`, `scikit-learn.Predictor`, `scikit-learn.Transformer`,
        and `scikit-learn.Model`.
    num_vars : `numpy.array` or list, default=None
        Array of numerical features. All variable in `num_vars` must be in `original_features`.
    cat_vars : `numpy.array` or list, default=None
        Array of categorical features. All variable in `cat_vars` must be in `original_features`.
    features_to_keep: list, default=[]
        In case some features are of special interest to the analysis and should be kept.
    {num,mixed,cat}_kws : dict, default=None
        Additional keyword arguments to pass to `compute_high_corr_pairs()`, `compute_highly_related_mixed_vars()`, and
        `compute_highly_related_categorical_vars()`.
    rfecv_kws : dictionary, default=None
        Dictionary for RFECV.

    Returns
    ----------
    filtered_features : list
        List with selected features.
    """

    tot_vars = 0
    if cat_vars:
        tot_vars += len(cat_vars)
    if num_vars:
        tot_vars += len(num_vars)
    assert len(original_features) == tot_vars, "`original_features` != `num_vars` + `cat_vars`"

    # First, we compute highly related pairs of variables
    if num_kws is None:
        num_kws = dict(corr_thres=0.8)
    hi_rel = _compute_highly_related_pairs(df, num_vars, cat_vars, num_kws, mixed_kws, cat_kws)

    filtered_features = original_features.copy()

    if hi_rel.shape[0] > 0:
        # Next, we remove variables highly correlated with the ones indicated to be kept
        for v in features_to_keep:
            for v2 in hi_rel.loc[hi_rel['var1'] == v, 'var2'].to_list():
                if v2 in filtered_features:
                    if v2 in features_to_keep:
                        logger.warning(
                            f'Variables {v} and {v2} are highly correlated, and both were selected to be kept.')
                    else:
                        filtered_features.remove(v2)

        # We update the highly-correlated pair DataFrame for efficiency purposes
        hi_rel = _compute_highly_related_pairs(df, num_vars, cat_vars, num_kws, mixed_kws, cat_kws)

    if hi_rel.shape[0] > 0:
        # Next, we remove highly correlated variables iteratively, keeping the most impactful one
        # We identify the most impactful one using a classification algorithm and SHAP importances
        # In every iteration, we remove all variables with a high correlation to the unvisited variable
        # with the highest SHAP importance value
        stop = False
        visited = []
        while not stop:
            X = df[filtered_features]
            y = target
            classifier.fit(X, y)
            importances = shap_importances(classifier, X)
            updated = False
            for feat in importances['variable_name']:
                if feat in hi_rel['var1'].to_list() and feat not in visited:
                    visited.append(feat)
                    for var2 in hi_rel.loc[hi_rel['var1'] == feat, 'var2'].to_list():
                        if var2 in filtered_features and var2 not in features_to_keep:
                            filtered_features.remove(var2)
                    updated = True
                    break

            stop = not updated

    # Finally, we apply Recursive Feature Elimination with Cross-Validation (RFECV)
    X = df[filtered_features]
    y = target
    if rfecv_kws is None:
        rfecv_kws = dict(step=1, cv=3)
    selector = RFECV(classifier, **rfecv_kws)
    selector = selector.fit(X, y)
    filtered_features = list(np.array(filtered_features)[np.where(selector.support_)])
    return filtered_features


def hyperparameter_tuning(X_train, y_train, classifier, param_grid, gridsearch_kws=None):
    """
    Runs grid search with cross-validation for hyperparameter tuning.

    Parameters
    ----------
    X_train : `np.ndarray` or `pandas.DataFrame`
        Training set.
    y_train : `numpy.array` or list
        Training target values.
    classifier : estimator, default=None
        Should be an instance of `scikit-learn.Estimator`, `scikit-learn.Predictor`, `scikit-learn.Transformer`,
        and `scikit-learn.Model`.
        If none is passed a `sklearn.ensemble.RandomForestClassifier` model is used.
    param_grid : dictionary
        Dictionary with the different hyperparameters and their corresponding values to be evaluated.
    gridsearch_kws : dictionary, default=None
        Dictionary for grid search.

    Returns
    ----------
    grid_result : `sklearn.model_selection.GridSearchCV`
        Instance of fitted estimator.
    """
    if gridsearch_kws is None:
        gridsearch_kws = dict(n_jobs=-1, cv=3)
    grid_search = GridSearchCV(classifier, param_grid, **gridsearch_kws)
    grid_result = grid_search.fit(X_train, y_train)
    return grid_result



"""Classification class"""
# Author: Miguel Alvarez-Garcia

import logging
import numpy as np
import utils

from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.model_selection import train_test_split
from xgboost import XGBClassifier

from .viz_utils import *


class Classifier:
    """
    Class to manage the classification model.

    Parameters
    ----------
    df : `pandas:DatasFrame`
        DataFrame with main data
    predictor_cols : list
        List of columns to use as predictors.
    target : `numpy.array` or list
        values of the target variable.
    num_cols : list, default=None
        List of numerical columns to use as predictors.
    cat_cols : list, default=None
        List of categorical columns to use as predictors.
    """

    def __init__(self, df, predictor_cols, target, num_cols=None, cat_cols=None):
        self.df = df
        self.original_features = predictor_cols
        self.num_vars = num_cols
        self.cat_vars = cat_cols
        self.filtered_features_ = None
        self.target = target
        self.model_ = None
        self.X_train_, self.X_test_, self.y_train_, self.y_test_ = tuple([None]*4)
        self.grid_result_ = None

        # Initialize logger
        logging.basicConfig(format='%(asctime)s %(levelname)-8s %(name)s: %(message)s',
                            level=logging.INFO,
                            datefmt='%Y-%m-%d %H:%M:%S')

        self.logger = logging.getLogger(__name__)

    def train_model(self, model=None, feature_selection=True, features_to_keep=[],
                    feature_selection_model=None, hyperparameter_tuning=False, param_grid=None,
                    train_size=0.8):
        """
        This method trains a classification model.

        By default, it uses XGBoost, but any other estimator (instance of `scikit-learn.Estimator`) can be used.

        The building process consists of four steps:
         - Train-test split
         - Feature Selection (optional)
           Feature removing highly correlated variables using a classification model and SHAP values
           to determine which to keep, and Recursive Feature Elimination with Cross-Validation (RFECV)
           on the remaining features.
         - Hyperparameter tuning (optional)
           Runs grid search with cross-validation for hyperparameter tuning. **Note** the parameter grid
           must be passed.
         - Model training
           Trains a classification model with the selected features and hyperparameters. By default, an XGBoost
           classifier will be trained.

        **Note** both hyperparameter tuning and model training are run on a train set. Train-test split is performed
        using `sklearn.model_selection.train_test_split`.

        Parameters
        ----------
        model : `scikit-learn.Estimator`, default=None
            Model to use as classifier. By default, `xgboost.XGBClassifier`.
        feature_selection : boolean, default=True
            If True, feature selection is performed on the original features.
        features_to_keep : list, default=empty list
            Features to be kept during the feature selection process.
        feature_selection_model : `scikit-learn.Estimator`, default=None
            Model to be used for feature selection. By default, `sklearn.ensemble.RandomForestClassifier`.
        hyperparameter_tuning : boolean, default=False
            If True, hyperparameter tuning is run using the parameter grid `param_grid`.
        param_grid : dictionary, default=None
            Parameter grid for hyperparameter tuning. **Note** if hyperparameter_tuning=True, a parameter grid is
            needed.
        train_size : float, default=0.8
            Size of the train split of the data for model training.
        """
        # Data is split into train and test sets
        X = self.df[self.original_features]
        y = self.target
        self.X_train_, self.X_test_, self.y_train_, self.y_test_ = train_test_split(X, y, train_size=train_size)

        # Feature selection
        if feature_selection:
            self.logger.info('Running feature selection...')
            if feature_selection_model is None:
                min_samples_leaf = int(np.ceil(self.X_train_.shape[0] * 0.05))
                feature_selection_model = RandomForestClassifier(max_depth=10,  min_samples_leaf=min_samples_leaf,
                                                                 random_state=42)

            self.filtered_features_ = utils.feature_selection(self.df.loc[self.X_train_.index], self.original_features,
                                                              self.target.loc[self.X_train_.index],
                                                              feature_selection_model, self.num_vars, self.cat_vars,
                                                              features_to_keep)

            self.X_train_ = self.X_train_[self.filtered_features_]
            self.X_test_ = self.X_test_[self.filtered_features_]
        else:
            self.filtered_features_ = self.original_features.copy()

        # Model instantiation
        if model is None:
            model = XGBClassifier(eval_metric='auc', use_label_encoder=False, random_state=42)
        self.model_ = model

        # Hyperparameter tuning
        if hyperparameter_tuning:
            if param_grid is None:
                raise RuntimeError('For hyperparameter tuning, some parameter grid must be passed - `param_grid`')
            self.logger.info('Running hyperparameter tuning...')
            self.grid_result_ = utils.hyperparameter_tuning(self.X_train_, self.y_train_, self.model_, param_grid)
            self.model_.set_params(**self.grid_result_.best_params_)

        # Model training
        self.logger.info('Training model...')
        self.model_.fit(self.X_train_, self.y_train_)
        self.logger.info('DONE!')

    @property
    def feature_importances(self):
        return utils.shap_importances(self.model_, self.X_train_)

    def hyperparameter_tuning_metrics(self, output_path=None):
        """
        This method returns the average and standard deviation of the cross-validation runs for every hyperparameter
        combination in hyperparameter tuning.

        Parameters
        ----------
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.
        """
        htm = pd.DataFrame(self.grid_result_.cv_results_['params'])
        htm.columns = pd.MultiIndex.from_product([[f'{self.model_.__class__.__name__} Hyperparameters'], htm.columns])
        htm[('Performance metrics', 'mean_test_score')] = self.grid_result_.cv_results_['mean_test_score']
        htm[('Performance metrics', 'std_test_score')] = self.grid_result_.cv_results_['std_test_score']

        if output_path is not None:
            htm.to_csv(output_path, index=False)

        return htm

    def confusion_matrix(self, test=True, sum_stats=True, output_path=None):
        """
        This method returns the confusion matrix of the classification model.

        Parameters
        ----------
        test : boolean, default=True
            If True, returns the confusion matrix calculated on the test set. If False, returns the confusion matrix on
            the train set.
        sum_stats : boolean, default=True
            If True, adds row (recall), column (precision), and global accuracy metrics to the matrix.
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        cm : `pandas.DataFrame`
            DataFrame with confusion matrix.
        """
        X = self.X_test_ if test else self.X_train_
        y = self.y_test_ if test else self.y_train_

        cm = pd.DataFrame(confusion_matrix(y, self.model_.predict(X)),
                          columns=pd.MultiIndex.from_product([['Predicted values'], np.unique(self.target)]),
                          index=pd.MultiIndex.from_product([['Observed values'], np.unique(self.target)]))

        if sum_stats:
            # Precision, recall, and global accuracy are appended to the table
            recall = np.diag(cm) / cm.sum(1)
            precision = np.diag(cm) / cm.sum()
            accuracy = np.diag(cm).sum() / cm.sum().sum()
            cm['recall'] = recall
            cm = cm.transpose()
            cm['precision'] = list(precision) + [accuracy]
            cm = cm.transpose()

        if output_path is not None:
            cm.to_csv(output_path, index=False)

        return cm

    def classification_report(self, test=True, output_path=None):
        """
        This method returns the `sklearn.metrics.classification_report` in `pandas.DataFrame` format.
        This report contains the intra-class metrics precision, recall and F1-score, together with the global accuracy,
        and macro average and weighted average of the three intra-class metrics.

        Parameters
        ----------
        test : boolean, default=True
            If True, returns the confusion matrix calculated on the test set. If False, returns the confusion matrix on
            the train set.
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        report : `pandas.DataFrame`
            DataFrame with classification report.
        """
        X = self.X_test_ if test else self.X_train_
        y = self.y_test_ if test else self.y_train_

        report = pd.DataFrame(classification_report(y, self.model_.predict(X), output_dict=True)).transpose()

        if output_path is not None:
            # In this case we do want to keep the index
            report.to_csv(output_path)

        return report

    def plot_shap_importances(self, n_top=7, output_path=None, savefig_kws=None):
        """
        Plots shap importance values, calculated as the combined average of the absolute values of the shap values
        for all classes.

        Parameters
        ----------
        n_top : int, default=7
           Top n features to be displayed. The importance of the rest are aggregated and displayed under the tag "Rest".
        output_path : str, default=None
           Path to save figure as image.
        savefig_kws : dict, default=None
           Save figure options.
        """
        plot_shap_importances(self.model_, self.X_train_, n_top=n_top, output_path=output_path, savefig_kws=savefig_kws)

    def plot_shap_importances_beeswarm(self, class_id, n_top=10, output_path=None, savefig_kws=None):
        """
        Plots a summary of shap values for a specific class of the target variable. This uses shap beeswarm plot
        (https://shap.readthedocs.io/en/latest/example_notebooks/api_examples/plots/beeswarm.html).

        Parameters
        ----------
        class_id : int
            The class for which to show the SHAP values.
        n_top : int, default=7
            Top n features to be displayed. The importances of the rest are aggregated and displayed under the tag "Rest".
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        plot_shap_importances_beeswarm(self.model_, self.X_train_, class_id, n_top=n_top, output_path=output_path,
                                       savefig_kws=savefig_kws)

    def plot_confusion_matrix(self, test=True, sum_stats=True, output_path=None, savefig_kws=None):
        """
        This function makes a pretty plot of an sklearn Confusion Matrix cf using a Seaborn heatmap visualization.

        Parameters
        ---------
        test : boolean, default=True
            If True, returns the confusion matrix calculated on the test set. If False, returns the confusion matrix on
            the train set.
        sum_stats : boolean, default=True
            If True, show precision and recall per class, and global accuracy, appended to the matrix.
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        cm = self.confusion_matrix(test, sum_stats)
        plot_confusion_matrix(cm, sum_stats=sum_stats, figsize=(cm.shape[0]+1, cm.shape[0]),
                              output_path=output_path, savefig_kws=savefig_kws)

    def plot_roc_curves(self, test=True, output_path=None, savefig_kws=None):
        """
       Plots ROC curve for every class.

       Parameters
       ---------
       test : boolean, default=True
            If True, returns the confusion matrix calculated on the test set. If False, returns the confusion matrix on
            the train set.
       output_path : str, default=None
           Path to save figure as image.
       savefig_kws : dict, default=None
           Save figure options.
       """
        X = self.X_test_ if test else self.X_train_
        y = self.y_test_ if test else self.y_train_
        plot_roc_curves(X, y, self.model_, output_path=output_path, savefig_kws=savefig_kws)



"""Table utils for dimensionality reduction"""
# Author: Miguel Alvarez-Garcia

import numpy as np
import pandas as pd

from ..utils import cross_corr_ratio


def cross_corr(df1, df2):
    """
    Calculates the correlation coefficient of every column in df1 with every column in df2

    Parameters
    ----------
    df1 : `pandas.DataFrame`
    df2 : `pandas.DataFrame`

    Returns
    ----------
    corr_df: `pandas.DataFrame`
        DataFrame with the correlation coefficient of every pair of columns from df1 and df2.
    """
    corr_coeffs = []
    for col in df1.columns:
        corr_coeffs.append(df2.corrwith(df1[col]).tolist())

    corr_df = pd.DataFrame(corr_coeffs, index=df1.columns, columns=df2.columns).transpose()
    return corr_df


def num_main_contributors(df, df_trans, thres=0.5, n_contributors=None, dim_idx=None, component_description=None,
                          col_description=None, output_path=None):
    """
    Computes the original numerical variables with the strongest relation to the derived variable(s)
    (measured as Pearson correlation coefficient)

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame with original numerical variables.
    df_trans : `pandas.DataFrame`
        DataFrame with derived variables.
    thres : float, default=0.5
        Correlation coefficient threshold to consider one original variable to be a main contributor of a derived
        variable.
    n_contributors : float, default=None
        If n_contributors is passed, the n_contributors original variables with the highest correlation coefficient
        are selected for every derived variable.
        If n_contributors is passed, the correlation coefficient threshold (thres) is ignored.
    dim_idx : int, default=None
        In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
    component_description: str or list
        Description of derived variables. It might be of interest to show a description of the new variables
        on a table for explainability purposes.
    col_description : `pandas.DataFrame`
        DataFrame with two columns: First one with original variable names, and a second one with the description.
        This is also used for explainability purposes.
    output_path : str, default=None
        If an output_path is passed, the resulting DataFame is saved as a CSV file.

    Returns
    ----------
    mc : `pandas.DataFrame`
        DataFrame with the main contributors of every derived variable.
    """
    if dim_idx is not None:
        df_trans = df_trans[df_trans.columns[dim_idx]].to_frame()

    if n_contributors is not None:
        thres = 0

    corrs = cross_corr(df, df_trans)
    mc = pd.DataFrame()
    for idx, row in corrs.iterrows():
        pc_corrs = row[(row < -thres) | (row > thres)].to_frame().reset_index().rename(
            columns={'index': 'var_name', idx: 'corr_coeff'})
        pc_corrs['component'] = idx
        pc_corrs['corr_coeff_abs'] = np.abs(pc_corrs['corr_coeff'])
        pc_corrs = pc_corrs.sort_values('corr_coeff_abs', ascending=False).drop(columns='corr_coeff_abs')

        if n_contributors is not None:
            pc_corrs = pc_corrs.iloc[:n_contributors]

        mc = pd.concat([mc, pc_corrs], ignore_index=True)

    final_cols = ['component', 'var_name', 'corr_coeff']

    if component_description is not None:
        if not isinstance(component_description, list):
            component_description = [component_description]

        component_description = pd.DataFrame(data={'component': mc['component'].unique(),
                                                   'component_description': component_description})
        final_cols.insert(1, 'component_description')
        mc = mc.merge(component_description, on='component')

    if col_description is not None:
        if not isinstance(col_description, pd.DataFrame):
            raise RuntimeError('`col_description` must be an instance of pandas.DataFrame')

        col_description.columns = ['var_name', 'var_description']
        final_cols.insert(final_cols.index('var_name')+1, 'var_description')
        mc = mc.merge(col_description, on='var_name')

    if output_path is not None:
        mc[final_cols].to_csv(output_path, index=False)

    return mc[final_cols]


def cat_main_contributors(df, df_trans, thres=0.14, n_contributors=None, dim_idx=None, component_description=None,
                          col_description=None, output_path=None):
    """
    Computes the original categorical variables with the strongest relation to the derived variable(s)
    (measured as correlation ratio)

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame with original categorical variables.
    df_trans : `pandas.DataFrame`
        DataFrame with derived variables.
    thres : float, default=0.14
        Correlation ratio threshold to consider one original variable to be a main contributor of a derived
        variable.
    n_contributors : float, default=None
        If n_contributors is passed, the n_contributors original variables with the highest correlation ratio
        are selected for every derived variable.
        If n_contributors is passed, the correlation ratio threshold (thres) is ignored.
    dim_idx : int, default=None
        In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
    component_description : str or list
        Description of derived variables. It might be of interest to show a description of the new variables
        on a table for explainability purposes.
    col_description : `pandas.DataFrame`
        DataFrame with two columns: First one with original variable names, and a second one with the description.
        This is also used for explainability purposes.
    output_path : str, default=None
        If an output_path is passed, the resulting DataFame is saved as a CSV file.

    Returns
    ----------
    mc : `pandas.DataFrame`
        DataFrame with the main contributors of every derived variable.
    """
    if dim_idx is not None:
        df_trans = df_trans[df_trans.columns[dim_idx]].to_frame()

    if n_contributors is not None:
        thres = 0

    corrs = cross_corr_ratio(df, df_trans)
    mc = pd.DataFrame()
    for idx, row in corrs.iterrows():
        pc_corrs = row[row > thres].to_frame().reset_index().rename(columns={'index': 'var_name', idx: 'corr_ratio'})
        pc_corrs['component'] = idx
        pc_corrs = pc_corrs.sort_values('corr_ratio', ascending=False).reset_index(drop=True)

        if n_contributors is not None:
            pc_corrs = pc_corrs.iloc[:n_contributors]

        mc = pd.concat([mc, pc_corrs], ignore_index=True)

    final_cols = ['component', 'var_name', 'corr_ratio']

    if component_description is not None:
        if not isinstance(component_description, list):
            component_description = [component_description]

        component_description = pd.DataFrame(data={'component': mc['component'].unique(),
                                                   'component_description': component_description})
        final_cols.insert(1, 'component_description')
        mc = mc.merge(component_description, on='component')

    if col_description is not None:
        if not isinstance(col_description, pd.DataFrame):
            raise RuntimeError('`col_description` must be an instance of pandas.DataFrame')

        col_description.columns = ['var_name', 'var_description']
        final_cols.insert(final_cols.index('var_name')+1, 'var_description')
        mc = mc.merge(col_description, on='var_name')

    if output_path is not None:
        mc[final_cols].to_csv(output_path, index=False)

    return mc[final_cols]


def cat_main_contributors_stats(df, df_trans, thres=0.14, n_contributors=None, dim_idx=None, output_path=None):
    """
    Computes for every categorical variable's value, the mean and std of the derived variables that are strongly
    related to the categorical variable (based on the correlation ratio)

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame with original categorical variables.
    df_trans : `pandas.DataFrame`
        DataFrame with derived variables.
    thres : float, default=0.14
        Correlation ratio threshold to consider one original variable to be a main contributor of a derived
        variable.
    n_contributors : float, default=None
        If n_contributors is passed, the n_contributors original variables with the highest correlation ratio
        are selected for every derived variable.
        If n_contributors is passed, the correlation ratio threshold (thres) is ignored.
    dim_idx : int, default=None
        In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
    output_path : str, default=None
        If an output_path is passed, the resulting DataFame is saved as a CSV file.

    Returns
    ----------
    stats : `pandas.DataFrame`
        DataFrame with the statistics.
    """
    if dim_idx is not None:
        df_trans = df_trans[df_trans.columns[dim_idx]].to_frame()

    m1 = pd.melt(df.reset_index().rename(columns={'index': 'id'}), id_vars=['id'], value_vars=df.columns)
    m2 = pd.melt(df_trans.reset_index().rename(columns={'index': 'id'}), id_vars=['id'], value_vars=df_trans.columns)
    stats = m1.merge(m2, on='id')
    stats = stats.groupby(['variable_y', 'variable_x', 'value_x']).agg({'value_y': ['mean', 'std']}) \
        .reset_index().rename(columns={'variable_y': 'component', 'variable_x': 'var_name', 'value_x': 'var_value'})
    stats.columns = [c[1] if c[1] != '' else c[0] for c in stats.columns]
    stats = stats.rename(columns={'mean': 'component_mean', 'std': 'component_std'})

    mc = cat_main_contributors(df, df_trans, thres=thres, n_contributors=n_contributors, dim_idx=dim_idx)
    stats = mc[['component', 'var_name']].merge(stats, on=['component', 'var_name'])

    if output_path is not None:
        stats.to_csv(output_path, index=False)

    return stats



"""Data preprocessing"""
# Author: Miguel Alvarez-Garcia

import networkx as nx
import numpy as np
import pandas as pd

from scipy.stats import rv_discrete
from sklearn.ensemble import IsolationForest
from sklearn.feature_selection import mutual_info_regression, mutual_info_classif
from sklearn.impute import KNNImputer
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mutual_info_score
from sklearn.preprocessing import MinMaxScaler
from ..utils import compute_high_corr_pairs, compute_highly_related_categorical_vars, compute_highly_related_mixed_vars


def compute_missing(df, normalize=True):
    """
    Calculates the pct/count of missing values per column.

    Parameters
    ----------
    df : `pandas.DataFrame`
    normalize : boolean, default=True

    Returns
    ----------
    missing_df : `pandas.DataFrame`
        DataFrame with the pct/counts of missing values per column.
    """
    missing_df = (df.isnull().sum()).to_frame('missing').reset_index().rename(columns={'index': 'var_name'})
    if normalize:
        missing_df['missing'] = missing_df['missing'] * 100 / df.shape[0]
    missing_df = missing_df.sort_values('missing', ascending=False)
    return missing_df


def num_imputation_pairs(df, corr_thres=0.7, method='pearson'):
    """
    Computes the correlation coefficient between every pair of variables and returns those pairs with an absolute value
    above the given threshold.
    *Note* all variables must be numerical.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    corr_thres : float, default=0.7
        Correlation theshold to consider two variables as strongly correlated.
    method : str, default='pearson'
        Method of correlation (pearson, kendall, spearman, or callable function -
        see https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.corr.html).

    Returns
    ----------
    num_pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables together with the correlation coefficient value.
    """
    return compute_high_corr_pairs(df, corr_thres=corr_thres, method=method)


def mixed_imputation_pairs(df, num_vars, cat_vars, np2_thres=0.14):
    """
    Computes the dependency between pairs of numerical and categorical variables through partial eta squared, and
    returns those pairs with a value above the given threshold.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    num_vars : string, list, series, or vector array
        Numerical variable name(s).
    cat_vars : string, list, series, or vector array
        Categorical variable name(s).
    np2_thres : float, default=0.14
        Threshold to consider two variables as strongly related (see
        https://www.spss-tutorials.com/effect-size/#anova-partial-eta-squared).

    Returns
    ----------
    pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables together with the partial eta squared value.
    """
    return compute_highly_related_mixed_vars(df, num_vars, cat_vars, np2_thres)


def cat_imputation_pairs(df, mi_thres=0.6):
    """
    Computes the dependency between paris of categorical variables through mutual information
    (https://en.wikipedia.org/wiki/Mutual_information), and returns those pairs with a value above the given threshold.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    mi_thres : float, default=0.6
        Threshold to consider two variables as strongly related.

    Returns
    ----------
    cat_pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables together with the mutual information score.
    """
    return compute_highly_related_categorical_vars(df, mi_thres)


def imputation_pairs(df, num_vars=None, cat_vars=None, num_kws=None, mixed_kws=None, cat_kws=None):
    """
    Computes one-to-one model-based imputation pairs. These are strongly related pairs of variables. Depending on the
    type of variables, a correlation coefficient (numerical variables), partial eta squared (mixed-type variables), or
    mutual information (categorical variables) is used to measured the strength of the relationship of each pair.

    When one variable is strongly related to more than one variable, the one that allows a larger amount of imputations.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    num_vars : str, list, series, or vector array
        Numerical variable name(s).
    cat_vars : str, list, series, or vector array
        Categorical variable name(s).
    {num,mixed,cat}_kws : dict, default=None
        Additional keyword arguments to pass to `num_imputation_pairs()`, `mixed_imputation_pairs()`, and
        `cat_imputation_pairs()`.

    Returns
    ----------
    final_pairs : `pandas.DataFrame`
        DataFrame with pairs of highly correlated variables (var1: variable with values to impute; var2: variable to be
        used as independent variable for model-based imputation), together proportion of missing values of variables
        var1 and var2.
    """
    if num_vars is None and cat_vars is None:
        raise ValueError('Numerical or categorical variable lists are required.')

    # Numerical variable pairs (correlation)
    num_pairs = pd.DataFrame()
    if num_vars:
        num_kws = num_kws if num_kws else dict()
        num_pairs = num_imputation_pairs(df[num_vars], **num_kws)

    # Mixed variable pairs (cross correlation)
    mixed_pairs = pd.DataFrame()
    if num_vars and cat_vars:
        mixed_kws = mixed_kws if mixed_kws else dict()
        mixed_pairs = mixed_imputation_pairs(df, num_vars, cat_vars, **mixed_kws)

    # Categorical variable pairs (mutual information)
    cat_pairs = pd.DataFrame()
    if cat_vars:
        cat_kws = cat_kws if cat_kws else dict()
        cat_pairs = cat_imputation_pairs(df[cat_vars], **cat_kws)

    final_pairs = pd.concat([num_pairs, mixed_pairs, cat_pairs], ignore_index=True)
    final_pairs = final_pairs.drop(columns='value')

    num_vars = num_vars if num_vars else []
    cat_vars = cat_vars if cat_vars else []

    # We only want to keep those pairs with missing values
    missing_df = compute_missing(df[num_vars + cat_vars])
    final_pairs = final_pairs.merge(missing_df, left_on='var1', right_on='var_name').rename(
        columns={'missing': 'missing_var1'}).drop(columns='var_name')
    final_pairs = final_pairs.merge(missing_df, left_on='var2', right_on='var_name').rename(
        columns={'missing': 'missing_var2'}).drop(columns='var_name')
    # Each variable with missing values and strongly related to at least another one, is assigned to the one from which
    # a larger proportion of missing values can be imputed.
    final_pairs = final_pairs.loc[final_pairs.groupby('var1')['missing_var2'].idxmin()].sort_values(
        ['missing_var2', 'missing_var1'])
    final_pairs = final_pairs[final_pairs['missing_var1'] > final_pairs['missing_var2']].reset_index(drop=True)
    return final_pairs


def _bucketize(values, q=10):
    """
    Quantile bucketizer.

    Parameters
    ----------
    values : `numpy.array`, list
        Array of values to bucketize.
    q : int, default=10
        Number of buckets.

    Returns
    ----------
    b_values : `pandas.Series`
        Series with bucketized values.
    """
    b_values = values
    if len(np.unique(values)) > 10:
        b_values = pd.qcut(values, q=q, duplicates='drop')
    return b_values


def impute_missing_values_with_highly_related_pairs(df_original, imputation_pairs, num_vars=None, cat_vars=None):
    """
    One-to-one model based imputation for the imputation pairs identified.
    * If both variables are numerical, linear regression is used. The regressor is trained using all observations with
    known values for both variables.
    * If both variables are categorical, we use hot deck imputation considering for every recipient all observations
    with the same value of the independent variable as donors. The value to be imputed is selected at random following
    the discrete empirical distribution.
    * For mixed-type variables, the numerical one is discretized by bucketization into quantiles and both variables are
    treated as categorical.

    Parameters
    ----------
    df_original : `pandas.DataFrame`
        DataFrame containing the data.
    imputation_pairs : `pandas.DataFrame`
        Imputation pairs as returned by the function `imputation_pairs()`.
    num_vars : str, list, series, or vector array
        Numerical variable name(s).
    cat_vars : str, list, series, or vector array
        Categorical variable name(s).

    Returns
    ----------
    df : `pandas.DataFrame`
        DataFrame with imputed data.
    """
    if num_vars is None and cat_vars is None:
        raise ValueError('Numerical and categorical variable lists are required.')

    num_vars = num_vars if num_vars else []
    cat_vars = cat_vars if cat_vars else []
    df = df_original.copy()

    for idx, row in imputation_pairs.iterrows():
        if row['var1'] in num_vars and row['var2'] in num_vars:
            # If both variables are numerical, use linear regression
            print(f"Imputing with linear regression {row['var2']} -> {row['var1']}")
            # First, model is fit for non-missing values
            lr = LinearRegression()
            df_f = df[(~df[row['var2']].isnull()) & (~df[row['var1']].isnull())]
            lr.fit(df_f[row['var2']].values.reshape(-1, 1), df_f[row['var1']])
            # Secondly, the fitted model is used for imputation
            df_f = df[(~df[row['var2']].isnull()) & (df[row['var1']].isnull())]
            df.loc[df_f.index, row['var1']] = lr.predict(df_f[row['var2']].values.reshape(-1, 1))

        else:
            print(f"Imputing with empirical discrete distribution {row['var2']} -> {row['var1']}")
            # Bucketize the numerical variable to treat both as categorical.
            if row['var2'] in num_vars:
                # print(f"Bucketize variable {row['var2']}")
                df[f"b_{row['var2']}"] = _bucketize(df[row['var2']])
                df[f"b_{row['var1']}"] = df[row['var1']]
            elif row['var1'] in num_vars:
                # print(f"Bucketize variable {row['var1']}")
                df[f"b_{row['var1']}"] = _bucketize(df[row['var1']])
                df[f"b_{row['var2']}"] = df[row['var2']]
            else:
                df[f"b_{row['var1']}"] = df[row['var1']]
                df[f"b_{row['var2']}"] = df[row['var2']]

            # Impute using the empirical discrete distribution function
            df_f = df[(~df[row['var2']].isnull()) & (df[row['var1']].isnull())]
            predictor_values = df_f[f"b_{row['var2']}"].unique().tolist()
            for pv in predictor_values:
                # For every value of the independent variable, we compute the discrete distribution frequencies of the
                # dependent variable calculated when both dependent and independent values are non-missing.
                dist_freq = df.loc[(df[f"b_{row['var2']}"] == pv) & (~df[row['var1']].isnull()), row['var1']]\
                    .value_counts(normalize=True).sort_index()
                if dist_freq.shape[0] > 0:
                    xk = tuple(dist_freq.index)
                    pk = tuple(dist_freq)
                    disc_dist = rv_discrete(name='imputation', values=(xk, pk))
                    # Impute missing values for the dependent variable using a random value extracted from the empirical
                    # discrete distribution computed above
                    no_missing = df[(df[f"b_{row['var2']}"] == pv) & (df[row['var1']].isnull())].shape[0]
                    df.loc[(df[f"b_{row['var2']}"] == pv) & (df[row['var1']].isnull()), row['var1']] = disc_dist.ppf(
                        np.random.rand(no_missing))

            df = df.drop(columns=[f"b_{row['var1']}", f"b_{row['var2']}"])

    return df


def mutual_information_pair_scores(df, num_vars=None, cat_vars=None):
    """
    Computes the mutual information score for every pair of variables.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    num_vars : str, list, series, or vector array
        Numerical variable name(s).
    cat_vars : str, list, series, or vector array
        Categorical variable name(s).

    Returns
    ----------
    mi_scores_df : `pandas.DataFrame`
        DataFrame with mutual information score for every pair of variables.
    """
    if num_vars is None and cat_vars is None:
        raise ValueError('Numerical and categorical variable lists are required.')

    num_vars = num_vars if num_vars else []
    cat_vars = cat_vars if cat_vars else []

    # First, we calculate max mutual information per variable for normalization
    max_mi_scores = dict()
    for nv in num_vars:
        values = df[nv].dropna().values
        mi = mutual_info_regression(values.reshape(-1, 1), values, discrete_features=False, random_state=42)[0]
        max_mi_scores[nv] = mi
    for cv in cat_vars:
        values = df[cv].dropna().values
        mi = mutual_info_score(values, values)
        max_mi_scores[cv] = mi

    # Secondly, we compute the mutual information of every pair of variables and normalize it by the max mutual
    # information calculated above
    mi_scores = list()
    # Numerical pairs
    for i in range(len(num_vars) - 1):
        for j in range(i + 1, len(num_vars)):
            df_f = df[[num_vars[i], num_vars[j]]].dropna()
            mi = max(mutual_info_regression(df_f[num_vars[i]].values.reshape(-1, 1), df_f[num_vars[j]],
                                            discrete_features=False, random_state=42)[0],
                     mutual_info_regression(df_f[num_vars[j]].values.reshape(-1, 1), df_f[num_vars[i]],
                                            discrete_features=False, random_state=42)[0])
            norm_factor = max(max_mi_scores[num_vars[i]], max_mi_scores[num_vars[j]])
            mi = mi / norm_factor
            mi_scores.append((num_vars[i], num_vars[j], mi))

    # Mixed pairs
    for i in range(len(num_vars)):
        for j in range(len(cat_vars)):
            df_f = df[[num_vars[i], cat_vars[j]]].dropna()
            mi = max(mutual_info_classif(df_f[num_vars[i]].values.reshape(-1, 1), df_f[cat_vars[j]],
                                         discrete_features=False, random_state=42)[0],
                     mutual_info_regression(df_f[cat_vars[j]].values.reshape(-1, 1), df_f[num_vars[i]],
                                            discrete_features=True, random_state=42)[0])
            norm_factor = max(max_mi_scores[num_vars[i]], max_mi_scores[cat_vars[j]])
            mi = mi / norm_factor
            mi_scores.append((num_vars[i], cat_vars[j], mi))

    # Categorical pairs
    for i in range(len(cat_vars) - 1):
        for j in range(i + 1, len(cat_vars)):
            df_f = df[[cat_vars[i], cat_vars[j]]].dropna()
            mi = mutual_info_score(df_f[cat_vars[i]], df_f[cat_vars[j]])
            norm_factor = max(max_mi_scores[cat_vars[i]], max_mi_scores[cat_vars[j]])
            mi = mi / norm_factor
            mi_scores.append((cat_vars[i], cat_vars[j], mi))

    mi_scores_df = pd.DataFrame(mi_scores, columns=['var1', 'var2', 'score'])
    return mi_scores_df


def variable_graph_partitioning(pair_scores, thres=0.05):
    """
    Computes a graph partitioning of the graph G=(V, E), where V = {variables} and E = {pairs of variables with a mutual
    information score above the threshold}. The connected components represent clusters of related variables.

    Parameters
    ----------
    pair_scores : `pandas.DataFrame`
        DataFrame with mutual information score for every pair of variables.
    thres : float, default=0.05
        Threshold to determine if two variables are similar based on mutual information score.

    Returns
    ----------
    all_nodes : list
        DataFrame with mutual information score for every pair of variables.
    edges : list of tuples
        List of graph edges
    g_list : list of sets
        List of connected components.
    """
    edges = pair_scores[pair_scores['score'] > thres].apply(lambda row: (row['var1'], row['var2']), axis=1).tolist()

    connected_nodes = list(zip(*edges))
    connected_nodes = list(connected_nodes[0]) + list(connected_nodes[1])
    nodes = list(np.unique(connected_nodes))

    G = nx.Graph()
    G.add_nodes_from(nodes)
    G.add_edges_from(edges)
    g_list = [c for c in sorted(nx.connected_components(G), key=len, reverse=True)]
    # Nodes, edges, connected components
    all_nodes = list(np.unique(pair_scores['var1'].to_list() + pair_scores['var2'].to_list()))
    return all_nodes, edges, g_list


def _hot_deck_weights(distances):
    """
    Callable function that selects a random neighbor for every missing value. This function accepts an array of
    distances, and returns an array of the same shape containing 0-1 weights.

    Parameters
    ----------
    distances : `numpy.ndarray`
        Array of distances of a recipient with its KNN neighbors.

    Returns
    ----------
    weights : `numpy.ndarray`
        Array with 0-1 weights.
    """
    weights = np.zeros(np.shape(distances))

    # Randomly select the neighbor that will be used to impute each of the missing values
    ids = np.random.randint(0, np.shape(distances)[1], size=np.shape(distances)[0])
    if np.shape(distances)[0] == 1:
        ids = [ids]

    i = 0
    for idx in ids:
        weights[i][idx] = 1
        i += 1

    return weights


def hot_deck_imputation(df_original, variables, k=8, partitions=None):
    """
    Computes KNN hot deck imputation using sklearn KNNImputer
    (see https://scikit-learn.org/stable/modules/generated/sklearn.impute.KNNImputer.html).

    Parameters
    ----------
    df_original : `pandas.DataFrame`
        DataFrame containing the data.
    variables : list
        List of variables with potential missing values.
        **Note** all partitions must be contained in this variable list.
    k : int, default=8
        Number of neighbors.
    partitions : `numpy.ndarray`, list of lists
        Variable partitions/clusters to impute variables by cluster.
        **Note** if any partition is passed, imputation is only performed on those observations with fewer than 1/4 of
        their values within the partition variables missed.

    Returns
    ----------
    df : `pandas.DataFrame`
        DataFrame with imputed values.
    """
    apply_threshold = True
    if partitions is None:
        apply_threshold = False
        partitions = [set(variables)]

    original_index = df_original.index
    df = df_original.sample(frac=1, random_state=42)

    mms = MinMaxScaler()
    df[variables] = mms.fit_transform(df[variables])

    for p in partitions:
        p = sorted(p)
        print('Partition:', p)
        imputer = KNNImputer(n_neighbors=k, weights=_hot_deck_weights)
        df_p = df
        if apply_threshold:
            df_p = df[df[p].isnull().sum(1) <= np.floor(len(p) / 4)]
        df_p_i = pd.DataFrame(data=imputer.fit_transform(df_p[p]), index=df_p.index, columns=p)
        df.loc[df_p_i.index, df_p_i.columns] = df_p_i.values
        print('Remaining missing values', df.isnull().sum().sum())

    df[variables] = mms.inverse_transform(df[variables])
    df = df.loc[original_index]
    return df


def impute_missing_values(df, num_vars, cat_vars, num_pair_kws=None, mixed_pair_kws=None, cat_pair_kws=None,
                          graph_thres=0.05, k=8, max_missing_thres=0.33):
    """
    Main function for data imputation. The procedure consists of four different steps:
     1. One-to-one model based imputation for strongly related variables (see
        `impute_missing_values_with_highly_related_pairs` doc).
     2. Cluster based hot deck imputation. Clusters of strongly related variables are computed using mutual information
        between pairs of variables and calculating the connected components of the undirected graph G=(V, E), where V is
        the set of variables and E the pairs of variables with mutual information above a predefined threshold.
     3. Records with a proportion of missing values above a predefined threshold are discarded to ensure the quality of
        the hot deck imputation.
     4. Hot deck imputation for the remaining missing values considering all variables together.

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    num_vars : str, list, series, or vector array
        Numerical variable name(s).
    cat_vars : str, list, series, or vector array
        Categorical variable name(s).
    {num,mixed,cat}_pair_kws : dict, default=None
        Additional keyword arguments to pass to compute imputation_pairs.
    graph_thres : float, default=0.05
        Threshold to determine if two variables are similar based on mutual information score, and therefore should be
        considered as edges of the graph from which variable clusters are derived.
    k : int, default=8
        Number of neighbors to consider in hot deck imputation.
    max_missing_thres: float, default=0.33
        Max proportion of missing values per record allowed before final general hot deck imputation.

    Returns
    ----------
    df : `pandas.DataFrame`
        DataFrame with all missing values imputed.
    """
    # One-to-one model based imputation for strongly related variables
    print('--- One-to-one model based imputation for strongly related variables ---')
    ip = imputation_pairs(df, num_vars, cat_vars, num_kws=num_pair_kws, mixed_kws=mixed_pair_kws, cat_kws=cat_pair_kws)
    df = impute_missing_values_with_highly_related_pairs(df, ip, num_vars, cat_vars)

    # Cluster based hot deck imputation
    # Applied only to clusters with at least 4 variables
    print('--- Cluster based hot deck imputation ---')
    mi_scores = mutual_information_pair_scores(df, num_vars, cat_vars)
    nodes, edges, comp = variable_graph_partitioning(mi_scores, thres=graph_thres)
    df = hot_deck_imputation(df, num_vars + cat_vars, k=k,
                             partitions=list(np.array(comp)[np.array(list(map(len, comp))) >= 4]))

    # Remove records with more than max_missing_thres missing values are removed.
    print(f'--- Removing records with more than {max_missing_thres*100} missing values ---')
    df = df[df.isnull().sum(1) < df.shape[1] * max_missing_thres]

    # Hot deck imputation for the remaining missing values
    print('--- Hot deck imputation for the remaining missing values ---')
    df = hot_deck_imputation(df, num_vars + cat_vars, k=k)
    return df


def remove_outliers(df, variables, iforest_kws=None):
    """
    Removes outliers using the Isolation Forest algorithm
    (https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.IsolationForest.html).

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    variables : list
        Variables with potential outliers.
    iforest_kws : dict, default=None
        IsolationForest algorithm hyperparameters.

    Returns
    ----------
    df_inliers : `pandas.DataFrame`
        DataFrame with inliers (i.e. observations that are not outliers).
    df_outliers : `pandas.DataFrame`
        DataFrame with outliers.
    """
    if iforest_kws is None:
        iforest_kws = dict(max_samples=0.8, max_features=0.8, bootstrap=False, random_state=42)
    outlier_if = IsolationForest(**iforest_kws)
    outlier_flag = outlier_if.fit_predict(df[variables])
    return df[outlier_flag > 0], df[outlier_flag < 0]



from .data_preprocessing import (
    compute_missing,
    hot_deck_imputation,
    imputation_pairs,
    impute_missing_values,
    impute_missing_values_with_highly_related_pairs,
    mutual_information_pair_scores,
    remove_outliers,
    variable_graph_partitioning
)

from .viz_utils import (
	missing_values_heatmap,
    plot_imputation_pairs_scatter,
    plot_imputation_distribution_assessment,
    plot_variable_graph_partitioning_components
)

__all__ = [
    # data processing
    "compute_missing",
    "hot_deck_imputation",
    "imputation_pairs",
    "impute_missing_values",
    "impute_missing_values_with_highly_related_pairs",
    "mutual_information_pair_scores",
    "remove_outliers",
    "variable_graph_partitioning",
    # visualization utils
	"missing_values_heatmap",
    "plot_imputation_pairs_scatter",
    "plot_imputation_distribution_assessment",
    "plot_variable_graph_partitioning_components"
]



"""Visualization utils for dimensionality reduction"""
# Author: Miguel Alvarez-Garcia

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from kneed import KneeLocator
from matplotlib.gridspec import GridSpec

from table_utils import cat_main_contributors, num_main_contributors
from ..utils import get_axis, plot_optimal_normalized_elbow, savefig

sns.set_style('whitegrid')

__types__ = ['cumulative', 'ratio', 'normalized']


def _plot_cumulative_explained_var(explained_variance_ratio, kl, thres, ax):
    n_components = len(explained_variance_ratio)

    ax.plot(np.append([0], explained_variance_ratio.cumsum()), label='', color='#332288')

    ax.axhline(thres, 0.01, 0.99, linestyle='--', linewidth=1, color='grey',
               label=f'{int(thres * 100)}% Explained variance')
    ax.set_yticks(np.append(ax.get_yticks()[1:-1], [thres]))

    ax.axvline(kl.knee, linestyle='--', linewidth=1, color='#E73F74', label=f'Optimal number of components')
    ax.axvline((explained_variance_ratio.cumsum() < thres).sum()+1, linestyle='--', linewidth=1, color='#11A579',
               label=f'Minimum number of components for {int(thres * 100)}% explained variance')
    ax.set_xticks(np.append(ax.get_xticks()[1:-1], [kl.knee, (explained_variance_ratio.cumsum() < thres).sum()+1]))
    ax.set_xlim(-n_components * 0.02, n_components * 1.02)

    ax.set_ylabel('Explained variance (cumulative ratio)', fontsize=13, labelpad=15)
    ax.legend(fontsize=12, labelspacing=0.5)


def _plot_explained_var_ratio(explained_variance_ratio, kl, ax):
    n_components = len(explained_variance_ratio)

    ax.plot([np.nan] + list(explained_variance_ratio), color='#332288', label='')

    avg_explained_var = 1 / (n_components - 1)
    ax.axhline(avg_explained_var, 0.01, 0.99, linestyle='--', linewidth=1, color='grey',
               label='Average explained variance (%)')
    ax.set_yticks(np.append(ax.get_yticks()[1:-1], [avg_explained_var]))

    ax.axvline(kl.knee, linestyle='--', linewidth=1, color='#E73F74', label=f'Optimal number of components')
    ax.axvline((explained_variance_ratio > avg_explained_var).sum(), linestyle='--', linewidth=1,
               color='#11A579', label=f'Number of components above average explained variance')
    ax.set_xticks(
        np.append(ax.get_xticks()[1:-1], [kl.knee, (explained_variance_ratio > avg_explained_var).sum()]))
    ax.set_xlim(-n_components * 0.02, n_components * 1.02)

    ax.set_ylabel('Explained variance (ratio)', fontsize=13, labelpad=8)
    ax.legend(fontsize=12, labelspacing=0.5)


def _plot_normalized_explained_var(explained_variance_ratio, kl, ax):
    plot_optimal_normalized_elbow(explained_variance_ratio, kl, ax, optimal_label='Optimal number of components',
                                  xlabel='Number of components', ylabel='Normalized explained variance curve')


def plot_explained_variance(explained_variance_ratio, thres=0.5, plots='all', output_path=None, savefig_kws=None):
    """
    Plot the explained variance (ratio, cumulative, and/or normalized)

    Parameters
    ----------
    explained_variance_ratio : `numpy.array`
        Array with the explained variance ratio of the first n components.
        Note it is assumed this is provided in descending order.
    thres : float, default=0.5
        Minimum explained cumulative variance ratio.
    plots : str or list, default='all'
        The following plots are supported: ['cumulative', 'ratio', 'normalized']
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """

    n_components = len(explained_variance_ratio)

    if plots == 'all':
        plots = __types__
    if not isinstance(plots, list):
        plots = [plots]

    kl = KneeLocator(x=range(1, n_components + 1), y=explained_variance_ratio, curve='convex',
                     direction='decreasing')

    fig, axs = plt.subplots(len(plots), 1, figsize=(8, 5 * len(plots)))

    i = 0
    for p in plots:

        ax = axs
        if len(plots) > 1:
            ax = axs[i]

        if p == 'cumulative':
            _plot_cumulative_explained_var(explained_variance_ratio, kl, thres, ax)
        elif p == 'ratio':
            _plot_explained_var_ratio(explained_variance_ratio, kl, ax)
        elif p == 'normalized':
            _plot_normalized_explained_var(explained_variance_ratio, kl, ax)
        else:
            raise NameError(f"Plot type '{p}' does not exists")
        i += 1

    plt.tight_layout(pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_num_main_contributors(df, df_trans, thres=0.5, n_contributors=5, dim_idx=None, output_path=None,
                               savefig_kws=None):
    """
    Plot main contributors (original variables with the strongest relation with derived variables) for
    every derived variable

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame with original numerical variables.
    df_trans : `pandas.DataFrame`
        DataFrame with derived variables.
    thres : float, default=0.5
        Minimum Pearson correlation coefficient to consider an original and a derived variable to be strongly related.
    n_contributors : int, default=5
        Number of contributors by derived variables (the ones with the strongest correlation coefficient
        are shown).
    dim_idx : int, default=None
        In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """

    cmap = matplotlib.cm.get_cmap('coolwarm')
    n_contributors = np.minimum(n_contributors, df.shape[1])
    mc = num_main_contributors(df, df_trans, thres=thres, n_contributors=n_contributors, dim_idx=dim_idx)
    mc['corr_coeff_abs'] = np.abs(mc['corr_coeff'])
    mc = mc.sort_values(by=['component', 'corr_coeff_abs']).reset_index(drop=True).drop(columns='corr_coeff_abs')
    nplots = mc['component'].nunique()
    ncols = 2
    if nplots == 1:
        ncols = 1
    if nplots % 2 > 0 or nplots % 3 == 0:
        ncols = 3

    nbars = mc.groupby('component')['var_name'].count().max()

    nrows = nplots // ncols + (nplots % ncols > 0)
    fig, axs = plt.subplots(nrows, ncols, figsize=(6 * ncols, 0.6 * nbars * nrows))
    xticks = (np.array(range(9)) - 4) / 4

    i = 0
    for pc in mc['component'].unique():
        ax = get_axis(i, axs, ncols, nrows)
        n_pc_contrib = mc[mc['component'] == pc].shape[0]
        ax.barh(y=range(n_pc_contrib), width=mc.loc[mc['component'] == pc, 'corr_coeff'],
                color=list(map(lambda x: cmap(x), (mc.loc[mc['component'] == pc, 'corr_coeff'] + 1) / 2)), alpha=0.95)
        ax.vlines(0, -0.5, n_pc_contrib - 0.5, color='black', linewidth=0.5)
        if i // ncols == nrows - 1:
            ax.set_xlabel('Correlation coefficient', fontsize=12)
        ax.set_xticks(ticks=xticks)
        ax.set_yticks(ticks=range(n_pc_contrib))
        ax.set_yticklabels(labels=mc.loc[mc['component'] == pc, 'var_name'], rotation=0, fontsize=11)
        ax.set_title(str.upper(pc), fontsize=13)
        i += 1

    while i < ncols*nrows:
        ax = get_axis(i, axs, ncols, nrows)
        ax.axis('off')
        i += 1

    fig.tight_layout(pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_cat_main_contributor_distribution(df, df_trans, thres=0.14, n_contributors=None, dim_idx=None,
                                           output_path=None, savefig_kws=None):
    """
    Plot main contributors (original variables with the strongest relation with derived variables) for
    every derived variable

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame with original categorical variables.
    df_trans : `pandas.DataFrame`
        DataFrame with derived variables.
    thres : float, default=0.5
         Minimum correlation ratio to consider an original and a derived variable to be strongly related.
    n_contributors: int, default=5
        Number of contributors by derived variables (the ones with the strongest correlation coefficient
        are shown).
    dim_idx : int, default=None
        In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    dim_name = ''
    if dim_idx is not None:
        dim_name = df_trans.columns[dim_idx]
        df_trans = df_trans[dim_name].to_frame()
    else:
        raise RuntimeWarning(
            '''`plot_disc_main_contributor_distribution` is designed to plot one component at a time. 
            Provide a value for dim_idx''')

    mc = cat_main_contributors(df, df_trans, thres=thres, n_contributors=n_contributors, dim_idx=dim_idx)

    ncols = mc.shape[0]
    if ncols > 5:
        raise Warning(f'''{ncols} original variables are highly related to de new construct variable.
                      Only the strongest 5 will be shown.''')
        ncols = 5

    nrows = np.lcm.reduce(df[mc['var_name'].tolist()].nunique().tolist())

    fig = plt.figure(figsize=(np.minimum(6 * ncols, 20), 5))
    gs = GridSpec(nrows, ncols, figure=fig)
    df = pd.concat([df, df_trans], axis=1)

    j = 0
    for idx, row in mc.iterrows():
        var_name = row['var_name']
        nvalues = df[var_name].nunique()

        i = 0
        ax0 = None
        for v in df[var_name].unique():
            ax = fig.add_subplot(gs[i:i + int(nrows / nvalues), j], sharex=ax0, sharey=ax0)
            sns.kdeplot(data=df[df[var_name] == v], x=dim_name, color='blue', fill=True, ax=ax)
            ax.set_title(f'{var_name} = {v}', fontsize=10)
            ax.set_ylabel('')

            i += int(nrows / nvalues)
            if i < nrows:
                ax.set_xlabel('')

            if ax0 is None:
                ax0 = ax
        j += 1

    sns.despine(fig)
    fig.supylabel('Density functions', fontsize=13, x=0.01)
    fig.tight_layout(w_pad=3)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_cumulative_explained_var_comparison(explained_variance_ratio1, explained_variance_ratio2, name1=None,
                                             name2=None, thres=None, output_path=None, savefig_kws=None):
    """
    Plots comparison of cumulative explained variance between two techniques.

    Parameters
    ----------
    explained_variance_ratio1 : list, `numpy.array`
        Explained variance ratio by technique 1.
    explained_variance_ratio2 : list, `numpy.array`
        Explained variance ratio by technique 2.
    name1 : str, default=None
        Name of technique 1. (For styling purposes).
    name2 : str, default=None
        Name of technique 2. (For styling purposes).
    thres : float, default=None
         Reference threshold for cumulative explained variance ratio. (For styling purposes).
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    fig, ax = plt.subplots(figsize=(7, 5))
    ax.plot(np.append([0], explained_variance_ratio1), color='#7F3C8D', label=name1)
    ax.plot(np.append([0], explained_variance_ratio2), color='#11A579', label=name2)
    if thres is not None:
        ax.axhline(thres, 0.01, 0.99, linestyle='--', linewidth=1, color='grey',
                   label=f'{int(thres*100)}% Explained variance')
        ax.set_yticks(np.append(ax.get_yticks()[1:-1], [thres]))
    n_components = np.maximum(len(explained_variance_ratio1), len(explained_variance_ratio2))
    ax.set_xlim(-n_components * 0.02, n_components * 1.02)
    ax.set_ylabel('Explained variance (cumulative ratio)', fontsize=13, labelpad=15)
    ax.legend(fontsize=12, labelspacing=0.5)
    ax.set_xlabel('Number of components', fontsize=13, labelpad=15)
    ax.legend(fontsize=12, labelspacing=0.5)
    fig.tight_layout(w_pad=2)
    savefig(output_path=output_path, savefig_kws=savefig_kws)


def plot_compare_pca_based_components(components_pca, components_other, original_vars, other_name='Sparse PCA', n_pc=1,
                                      output_path=None, savefig_kws=None):
    """
    Plots comparison of cumulative explained variance between two techniques.

    Parameters
    ----------
    components_pca : `numpy.array`
        Components of the `n_pc` principal component calculated with PCA.
    components_other : `numpy.array`
        Components of the `n_pc` principal component calculated a PCA based technique.
    original_vars : list
        List of names of the original variables.
    other_name : str, default='Sparse PCA'
        Name of the other applied technique.
    n_pc : int, default=1
        Principal component to which the components refer to.
    output_path : str, default=None
        Path to save figure as image.
    savefig_kws : dict, default=None
        Save figure options.
    """
    plt.figure(figsize=(16, 5))
    plt.bar(x=range(len(components_pca)), height=components_pca, label='PCA', color='#11A579', alpha=0.5)
    plt.plot(range(len(components_other)), components_other, 'o', label=other_name, color='#CF1C90', alpha=0.67)

    plt.hlines(0, 0, len(components_other), color='black', linewidth=0.5)
    plt.ylabel('Coefficients', fontsize=12)
    plt.xticks(ticks=range(len(original_vars)), labels=original_vars, rotation=90, fontsize=11)
    plt.legend(fontsize=12, title='Method', title_fontsize=13, labelspacing=0.5)
    plt.xlim(-1, len(original_vars))
    plt.title(f'Principal Component {str(n_pc).zfill(2)}', fontsize=14)
    plt.tight_layout()
    savefig(output_path=output_path, savefig_kws=savefig_kws)



from setuptools import setup, find_packages
import os

ROOT = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(ROOT, 'README.md')) as f:
    README = f.read()


setup(
    name='clust-learn',
    version=1.0,
    description="A Python package for explainable cluster analysis",
    long_description=README,
    long_description_content_type="text/markdown",
    author="Miguel Alvarez-Garcia, Raquel Ibar-Alonso, Mar Arenas-Parra",
    author_email="@gmail.com",
    url="https://github.com/malgar/clust-learn",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        'Topic :: Scientific/Engineering',
        'Programming Language :: Python :: 3.9'
    ],
    license='GPLv3',
    install_requires=[
        "kneed==0.7.0",
        "matplotlib==3.4.3",
        "networkx==2.6.3",
        "numpy==1.20.3",
        "pandas==1.3.4",
        "pingouin==0.5.1",
        "prince==0.7.0",
        "scikit-learn==1.0.2",
        "scipy==1.7.1",
        "seaborn==0.11.2",
        "shap==0.40.0",
        "statsmodels==0.13.2",
        "xgboost==1.5.2"
    ]
)



"""Dimensionality Reduction with PCA as reference"""
# Author: Miguel Alvarez-Garcia

import prince

from sklearn.decomposition import PCA, SparsePCA
from sklearn.preprocessing import StandardScaler
from .table_utils import *
from .viz_utils import *


class DimensionalityReduction:
    """
    Dimensionality Reduction class

    Parameters
    ----------
    df : `pandas.DataFrame`
        DataFrame containing the data.
    num_vars : string, list, series, or vector array
        Numerical variable name(s).
    cat_vars : string, list, series, or vector array
        Categorical variable name(s).
    num_algorithm : string, default='pca'
        Technique to be used for dimensionality reduction for numerical variables.
        By default, PCA (Principal Component Analysis) is used.
    cat_algorithm : string, default='mca'
        Technique to be used for dimensionality reduction for categorical variables.
        By default, MCA (Multiple Correspondence Analysis) is used.
    num_kwargs : dictionary, default=None
        Additional keyword arguments to pass to the model used for numerical variables.
    cat_kwargs : dictionary, default=None
        Additional keyword arguments to pass to the model used for categorical variables.
    """

    def __init__(
            self,
            df,
            num_vars=None,
            cat_vars=None,
            num_algorithm='pca',
            cat_algorithm='mca',
            num_kwargs=None,
            cat_kwargs=None
    ):
        self.num_algorithm = str.lower(num_algorithm)
        self.cat_algorithm = str.lower(cat_algorithm)
        self.n_components_ = None
        self.min_explained_variance_ratio_ = 0.5
        self.df = df
        self.num_vars = num_vars
        self.cat_vars = cat_vars
        self.num_trans_ = None
        self.cat_trans_ = None
        self.num_components_ = None
        self.cat_components_ = None
        self.num_kwargs = {} if num_kwargs is None else num_kwargs
        self.cat_kwargs = {} if cat_kwargs is None else cat_kwargs

        # Regardless of the algorithm selected for numerical variables, PCA is used as reference algorithm
        self.pca_ = PCA()

        self.num_model = None
        if self.num_algorithm == 'pca':
            self.num_model = PCA(n_components=None, random_state=42, **self.num_kwargs)
        elif self.num_algorithm in ['spca', 'sparsepca']:
            self.num_model = SparsePCA(n_components=None, random_state=42, **self.num_kwargs)
        else:
            raise RuntimeError('''An error occurred while initializing the algorithm.
                               Check the algorithm name for numerical variables.''')

        self.cat_model = None
        if self.cat_algorithm == 'mca':
            self.cat_model = prince.MCA(n_components=None, n_iter=10, random_state=42, **self.cat_kwargs)
        else:
            raise RuntimeError('''An error occurred while initializing the algorithm.
                               Check the algorithm name for categorical variables.''')

    def transform(self, n_components=None, min_explained_variance_ratio=0.5):
        """
        Transforms a DataFrame df to a lower dimensional space

        Parameters
        ----------
        n_components : int, default=None
            Number components to compute. If None, then `n_components` is set to the number of features.
            Note this number is approximate because numerical and categorical vars are treated independently.
        min_explained_variance_ratio : float, default=0.5
            Minimum explained variance ratio to be achieved. If `n_components` is not None,
            `min_explained_variance_ratio` will be ignored.
            If None: optimal

        Returns
        ----------
        trans : `pandas.DataFrame`
            DataFrame with the transformed data.
        """

        self.n_components_ = n_components
        self.min_explained_variance_ratio_ = min_explained_variance_ratio

        # If the number of components is specified and there are numerical and categorical variables
        # the number of components for each type of variables is distributed proportionally
        n_components_num = None
        n_components_cat = None
        if self.n_components_ is not None:
            if self.num_vars is None:
                n_components_cat = np.minimum(self.n_components_, self.df[self.cat_vars].nunique().sum())
            if self.cat_vars is None:
                n_components_num = np.minimum(self.n_components_, len(self.num_vars))
            if self.num_vars is not None and self.cat_vars is not None:
                n_components_num = int(
                    np.ceil(self.n_components_ * len(self.num_vars) / (len(self.num_vars) + len(self.cat_vars))))
                n_components_cat = int(
                    np.ceil(self.n_components_ * len(self.cat_vars) / (len(self.num_vars) + len(self.cat_vars))))

        trans = pd.DataFrame()
        if self.num_vars is not None:
            self.num_trans_ = self._transform_num(self.df[self.num_vars], n_components_num)
            trans = pd.concat([trans, self.num_trans_], axis=1)
        if self.cat_vars is not None:
            self.cat_trans_ = self._transform_cat(self.df[self.cat_vars], n_components_cat)
            trans = pd.concat([trans, self.cat_trans_], axis=1)

        idx_positions = np.maximum(len(str(trans.shape[1])), 2)
        trans.columns = [f'dim_{str(i+1).zfill(idx_positions)}' for i in range(trans.shape[1])]
        self.num_components_ = list(trans.columns)[:self.num_trans_.shape[1]]
        self.num_trans_.columns = self.num_components_
        self.cat_components_ = list(trans.columns)[-self.cat_trans_.shape[1]:]
        self.cat_trans_.columns = self.cat_components_
        self.n_components_ = len(self.num_components_) + len(self.cat_components_)
        return trans

    def _transform_num(self, df, n_components_num=None):
        # For now, we take PCA as reference because SPCA results depend on the number of components
        sc = StandardScaler()
        
        if self.pca_.n_components is None:
            self.pca_.fit(sc.fit_transform(df))

            if n_components_num is None and self.min_explained_variance_ratio_ is None:
                # Optimal number
                kl = KneeLocator(x=range(1, df.shape[1] + 1),
                                 y=self.pca_.explained_variance_ratio_,
                                 curve='convex',
                                 direction='decreasing')
                n_components_num = kl.knee

            elif n_components_num is None:
                # Based on explained variance
                n_components_num = (self.pca_.explained_variance_ratio_.cumsum() <
                                    self.min_explained_variance_ratio_).sum() + 1

            self.num_model.set_params(n_components=n_components_num)
            self.num_model.fit(sc.fit_transform(df))

        idx_positions = np.maximum(2, len(str(n_components_num)))
        trans = pd.DataFrame(self.num_model.transform(sc.fit_transform(df)),
                             columns=[f'dim_{str(i+1).zfill(idx_positions)}' for i in range(n_components_num)])
        
        # sort by explained variance
        trans = trans[pd.DataFrame(data={'pc': trans.columns, 'explained_var': trans.var().values})
                        .sort_values('explained_var', ascending=False)['pc']]
        trans.columns = [f'dim_{str(i+1).zfill(idx_positions)}' for i in range(trans.shape[1])]
        return trans

    def _transform_cat(self, df, n_components_cat=None):
        self.cat_model.set_params(n_components=df.nunique().sum())
        self.cat_model.fit(df.astype(str))

        explained_variance_ratio = self.cat_model.explained_inertia_ / self.cat_model.explained_inertia_.sum()
        if n_components_cat is None and self.min_explained_variance_ratio_ is None:
            # Optimal number
            kl = KneeLocator(x=range(1, df.nunique().sum() + 1),
                             y=explained_variance_ratio,
                             curve='convex',
                             direction='decreasing')
            n_components_cat = np.maximum(kl.knee-1, 1)

        elif n_components_cat is None:
            # Based on explained variance
            n_components_cat = (explained_variance_ratio.cumsum() <
                                self.min_explained_variance_ratio_).sum() + 1

        trans = self.cat_model.transform(df.astype(str))
        trans = trans[trans.columns[:n_components_cat]]
        idx_positions = np.maximum(2, len(str(n_components_cat)))
        trans.columns = [f'dim_{str(i+1).zfill(idx_positions)}' for i in range(n_components_cat)]
        return trans

    def num_main_contributors(self, thres=0.5, n_contributors=None, dim_idx=None, component_description=None,
                              col_description=None, output_path=None):
        """
        Computes the original numerical variables with the strongest relation to the derived variable(s)
        (measured as Pearson correlation coefficient)

        Parameters
        ----------
        thres : float, default=0.5
            Correlation coefficient threshold to consider one original variable to be a main contributor of a derived
            variable.
        n_contributors : float, default=None
            If n_contributors is passed, the n_contributors original variables with the highest correlation coefficient
            are selected for every derived variable.
            If n_contributors is passed, the correlation coefficient threshold (thres) is ignored.
        dim_idx : int, default=None
            In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
        component_description : str or list
            Description of derived variables. It might be of interest to show a description of the new variables
            on a table for explainability purposes.
        col_description : `pandas.DataFrame`
            DataFrame with two columns: First one with original variable names, and a second one with the description.
            This is also used for explainability purposes.
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        mc: `pandas.DataFrame`
            DataFrame with the main contributors of every derived variable.
        """
        return num_main_contributors(self.df[self.num_vars], self.num_trans_, thres, n_contributors, dim_idx,
                                     component_description, col_description, output_path)

    def cat_main_contributors(self, thres=0.14, n_contributors=None, dim_idx=None, component_description=None,
                              col_description=None, output_path=None):
        """
        Computes the original categorical variables with the strongest relation to the derived variable(s)
        (measured as correlation ratio)

        Parameters
        ----------
        thres : float, default=0.14
            Correlation ratio threshold to consider one original variable to be a main contributor of a derived
            variable.
        n_contributors : float, default=None
            If n_contributors is passed, the n_contributors original variables with the highest correlation ratio
            are selected for every derived variable.
            If n_contributors is passed, the correlation ratio threshold (thres) is ignored.
        dim_idx : int, default=None
            In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
        component_description : str or list
            Description of derived variables. It might be of interest to show a description of the new variables
            on a table for explainability purposes.
        col_description : `pandas.DataFrame`
            DataFrame with two columns: First one with original variable names, and a second one with the description.
            This is also used for explainability purposes.
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        mc: `pandas.DataFrame`
            DataFrame with the main contributors of every derived variable.
        """
        return cat_main_contributors(self.df[self.cat_vars], self.cat_trans_, thres, n_contributors, dim_idx,
                                     component_description, col_description, output_path)

    def cat_main_contributors_stats(self, thres=0.14, n_contributors=None, dim_idx=None, output_path=None):
        """
        Computes for every categorical variable's value, the mean and std of the derived variables that are strongly
        related to the categorical variable (based on the correlation ratio)

        Parameters
        ----------
        thres : float, default=0.14
            Correlation ratio threshold to consider one original variable to be a main contributor of a derived
            variable.
        n_contributors : float, default=None
            If n_contributors is passed, the n_contributors original variables with the highest correlation ratio
            are selected for every derived variable.
            If n_contributors is passed, the correlation ratio threshold (thres) is ignored.
        dim_idx : int, default=None
            In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
        output_path : str, default=None
            If an output_path is passed, the resulting DataFame is saved as a CSV file.

        Returns
        ----------
        stats: `pandas.DataFrame`
            DataFrame with the statistics.
        """
        return cat_main_contributors_stats(self.df[self.cat_vars], self.cat_trans_, thres, n_contributors, dim_idx,
                                           output_path)

    def plot_num_explained_variance(self, thres=0.5, plots='all', output_path=None, savefig_kws=None):
        """
        Plot the explained variance (ratio, cumulative, and/or normalized) for numerical variables

        Parameters
        ----------
        thres : float, default=0.5
            Minimum explained cumulative variance ratio.
        plots : str or list, default='all'
            The following plots are supported: ['cumulative', 'ratio', 'normalized']
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        plot_explained_variance(self.pca_.explained_variance_ratio_, thres, plots, output_path, savefig_kws)

    def plot_cat_explained_variance(self, thres=0.5, plots='all', output_path=None, savefig_kws=None):
        """
        Plot the explained variance (ratio, cumulative, and/or normalized) for categorical variables

        Parameters
        ----------
        thres : float, default=0.5
            Minimum explained cumulative variance ratio.
        plots : str or list, default='all'
            The following plots are supported: ['cumulative', 'ratio', 'normalized']
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        explained_variance_ratio = self.cat_model.explained_inertia_ / self.cat_model.explained_inertia_.sum()
        plot_explained_variance(explained_variance_ratio, thres, plots, output_path, savefig_kws)

    def plot_num_main_contributors(self, thres=0.5, n_contributors=5, dim_idx=None, output_path=None, savefig_kws=None):
        """
        Plot main contributors (original variables with the strongest relation with derived variables) for
        every derived variable

        Parameters
        ----------
        thres : float, default=0.5
            Minimum Pearson correlation coefficient to consider an original and a derived variable to be strongly
            related.
        n_contributors : int, default=5
            Number of contributors by derived variables (the ones with the strongest correlation coefficient
            are shown).
        dim_idx : int, default=None
            In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        plot_num_main_contributors(self.df[self.num_vars], self.num_trans_, thres, n_contributors, dim_idx, output_path,
                                   savefig_kws)

    def plot_cat_main_contributor_distribution(self, thres=0.14, n_contributors=None, dim_idx=None, output_path=None,
                                               savefig_kws=None):
        """
        Plot main contributors (original variables with the strongest relation with derived variables) for
        every derived variable

        Parameters
        ----------
        thres : float, default=0.5
             Minimum correlation ratio to consider an original and a derived variable to be strongly related.
        n_contributors : int, default=5
            Number of contributors by derived variables (the ones with the strongest correlation coefficient
            are shown).
        dim_idx : int, default=None
            In case only main contributors for derived variable in column position dim_idx are retrieved (starts at 0).
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        plot_cat_main_contributor_distribution(self.df[self.cat_vars], self.cat_trans_, thres, n_contributors, dim_idx,
                                               output_path, savefig_kws)

    def plot_cumulative_explained_var_comparison(self, thres=None, output_path=None, savefig_kws=None):
        """
        Plots comparison of cumulative explained variance between two techniques.

        Parameters
        ----------
        thres : float, default=None
             Reference threshold for cumulative explained variance ratio. (For styling purposes).
        output_path : str, default=None
            Path to save figure as image.
        savefig_kws : dict, default=None
            Save figure options.
        """
        plot_cumulative_explained_var_comparison(self.pca_.explained_variance_ratio_.cumsum(),
                                                 (self.num_trans_.var().values/len(self.num_vars)).cumsum(),
                                                 name1='pca', name2=self.num_algorithm, thres=thres,
                                                 output_path=output_path, savefig_kws=savefig_kws)
